import math
from typing import Union

from torch.utils.data import Dataset, random_split, DataLoader
import ancpbids


class TorchDataset(Dataset):
    def __init__(self, *bids_dataset_paths, **query_kwargs):
        self.data_events_files = []

        for bids_dataset_path in bids_dataset_paths:
            bids_dataset = ancpbids.load_dataset(bids_dataset_path)
            if query_kwargs is None:
                query_kwargs = {}
            # limit scope to raw as the user is most probably not interested in derivatives
            query_kwargs.setdefault("scope", "raw")
            data_files = bids_dataset.query(**query_kwargs)

            def load_contents(data_file) -> dict or None:
                sidecar = data_file.sidecar(suffix="events", extension=".tsv")
                if len(sidecar) > 0:
                    return sidecar[0].load_contents()
                return {}

            self.data_events_files += [
                (data_file.get_absolute_path(), load_contents(data_file)) for data_file in data_files]

    def __len__(self):
        return len(self.data_events_files)

    def __getitem__(self, idx):
        return self.data_events_files[idx]

    def split(self, *args: float):
        sum_ratios = sum(args)
        if not math.isclose(sum_ratios, 1.0):
            raise ValueError("sum of split ratios should be 1.0")

        return random_split(self, args)



import fnmatch
import inspect
import math
import os
import sys
from difflib import SequenceMatcher

from ancpbids.plugin import SchemaPlugin
from ancpbids.query import Select, query, query_entities
from ancpbids.utils import resolve_segments, convert_to_relative
from ancpbids.model_base import *


def has_entity(artifact, entity_):
    for e in artifact.entities:
        if e.key == entity_:
            return True
    return False


def get_entity(artifact, entity_):
    for e in artifact.entities:
        if e.key == entity_:
            return e.value
    return None


def add_entity(schema, artifact, key, value):
    if isinstance(key, schema.EntityEnum):
        key = key.entity_

    found = list(filter(lambda er: er.key == key, artifact.entities))
    if found:
        found[0].value = value
    else:
        eref = EntityRef(key, value)
        artifact.entities.append(eref)


def add_entities(schema, artifact, **kwargs):
    for k, v in kwargs.items():
        add_entity(schema, artifact, k, v)


def load_file_contents(folder, file_name, return_type: str = None):
    from ancpbids import utils
    file_path = get_absolute_path(folder, file_name)
    contents = utils.load_contents(file_path, return_type)
    return contents


def load_contents(file):
    from ancpbids import utils
    file_path = get_absolute_path(file.parent_object_, file.name)
    contents = utils.load_contents(file_path)
    return contents


def get_absolute_path_by_file(file):
    return get_absolute_path(file.parent_object_, file.name)


def get_absolute_path(folder, file_name=None):
    return _get_path(folder, file_name, True)


def _folder_get_relative_path(folder):
    return _get_path(folder, None, False)


def _file_get_relative_path(file):
    return _get_path(file.parent_object_, file.name, False)


def _get_path(folder, file_name=None, absolute=True):
    segments = []
    if file_name:
        segments.append(file_name)
    current_folder = folder
    while current_folder is not None:
        if isinstance(current_folder, Dataset):
            # prepend the dataset name if it is not part of the base dir name
            if not current_folder.base_dir_.endswith(current_folder.name):
                segments.insert(0, current_folder.name)
            if absolute:
                segments.insert(0, current_folder.base_dir_)
            # assume we reached the highest level, maybe not good for nested datasets
            break
        segments.insert(0, current_folder.name)
        current_folder = current_folder.parent_object_
    _path = os.path.join(*segments) if segments else ''
    _path = os.path.normpath(_path)
    if absolute:
        _path = os.path.abspath(_path)
    return _path


def remove_file(folder, file_name):
    folder.files = list(filter(lambda file: file.name != file_name, folder.files))


def create_artifact(folder, raw=None):
    artifact = Artifact()
    if isinstance(raw, Artifact):
        artifact.entities.extend(raw.entities)
    artifact.parent_object_ = folder
    folder.files.append(artifact)
    return artifact


def create_folder(folder, type_=None, **kwargs):
    if not type_:
        type_ = Folder
    sub_folder = type_(**kwargs)
    sub_folder.parent_object_ = folder
    folder.folders.append(sub_folder)
    return sub_folder


def create_derivative(ds, path=None, **kwargs):
    derivatives_folder = ds.derivatives
    if not ds.derivatives:
        derivatives_folder = DerivativeFolder()
        derivatives_folder.parent_object_ = ds
        derivatives_folder.name = "derivatives"
        ds.derivatives = derivatives_folder
    path = convert_to_relative(ds, path)
    target_folder, _ = resolve_segments(derivatives_folder, path, create_if_missing=True)
    derivative = DerivativeFolder(**kwargs)
    derivative.parent_object_ = target_folder
    target_folder.folders.append(derivative)

    derivative.dataset_description = DerivativeDatasetDescriptionFile()
    derivative.dataset_description.parent_object_ = derivative
    derivative.dataset_description.GeneratedBy = GeneratedBy()

    if ds.dataset_description:
        derivative.dataset_description.update(ds.dataset_description)

    return derivative


def create_dataset(schema, base_dir=None, **kwargs):
    ds = Dataset()
    ds._versioned_schema = schema
    ds.base_dir_ = base_dir
    ds.update(**kwargs)
    ds.dataset_description = DatasetDescriptionFile(name="dataset_description.json")
    ds.dataset_description.BIDSVersion = schema.VERSION
    ds.dataset_description.parent_object_ = ds
    return ds


def get_file(folder, file_name):
    folder, file_name = resolve_segments(folder, file_name, True)
    if not folder:
        return None
    direct_files = folder.to_generator(depth_first=True, depth=1, filter_=lambda n: isinstance(n, File))
    file = next(filter(lambda f: f.name == file_name, direct_files), None)
    return file


def get_files(folder, name_pattern):
    direct_files = folder.to_generator(depth_first=True, depth=1, filter_=lambda n: isinstance(n, File))
    return list(filter(lambda file: fnmatch.fnmatch(file.name, name_pattern), direct_files))


def remove_folder(folder, folder_name):
    folder.folders = list(filter(lambda f: f.name != folder_name, folder.folders))


def get_folder(folder, folder_name):
    direct_folders = folder.to_generator(depth_first=True, depth=1, filter_=lambda n: isinstance(n, Folder))
    return next(filter(lambda f: f.name == folder_name, direct_folders), None)


def get_files_sorted(folder):
    return sorted(folder.files, key=lambda f: f.name)


def get_folders_sorted(folder):
    return sorted(folder.folders, key=lambda f: f.name)


def to_generator(source, depth_first=False, filter_=None, depth=1000):
    if depth < 0:
        return

    if not depth_first:
        if filter_ and not filter_(source):
            return
        yield source

    for key, value in source.items():
        if isinstance(value, Model):
            yield from to_generator(value, depth_first, filter_, depth - 1)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, Model):
                    yield from to_generator(item, depth_first, filter_, depth - 1)

    if depth_first:
        if filter_ and not filter_(source):
            return
        yield source


def iterancestors(source):
    context = source
    while context is not None:
        if not hasattr(context, 'parent_object_'):
            break
        context = context.parent_object_
        yield context


def to_dict(source):
    return source


def get_model_classes(schema):
    if not hasattr(schema, '_CLASSES'):
        schema._CLASSES = {name: obj for name, obj in inspect.getmembers(schema) if inspect.isclass(obj)}
    return schema._CLASSES


def _get_element_members(schema, element_type):
    element_members = []
    try:
        members = element_type.MEMBERS
        element_members = list(
            map(lambda item: {'name': item[0], **item[1], 'type': _to_type(schema, item[1]['type'])},
                members.items()))
    except AttributeError as ae:
        pass
    return element_members


def get_members(schema, element_type, include_superclass=True):
    if element_type == schema.Model:
        return []
    super_members = []

    if include_superclass:
        superclass = element_type
        while True:
            try:
                mro = inspect.getmro(superclass)
                if len(mro) < 2:
                    break
                superclass = mro[1]
                if not superclass or superclass == schema.Model:
                    break
                super_members = super_members + _get_element_members(schema, superclass)
            except AttributeError:
                pass

    element_members = _get_element_members(schema, element_type)
    return super_members + element_members


def _to_type(schema, model_type_name: str):
    classes = schema.get_model_classes()
    if model_type_name in classes:
        return classes[model_type_name]
    if model_type_name in __builtins__:
        return __builtins__[model_type_name]
    return model_type_name


def _trim_int(value):
    try:
        # remove paddings/fillers in index values: 001 -> 1, 000230 -> 230
        # TODO return PaddedInt as done by PyBIDS
        return int(value)
    except ValueError:
        return value


def process_entity_value(schema, key, value):
    if not value:
        return value
    if isinstance(key, schema.EntityEnum):
        key = key.value['name']
    for sc_entity in filter(lambda e: e.value['name'] == key, schema.EntityEnum.__members__.values()):
        if sc_entity.value['format'] == 'index':
            if isinstance(value, list):
                return list(map(lambda v: _trim_int(v) if v is not None else v, value))
            else:
                return _trim_int(value)
    return value


def fuzzy_match_entity_key(schema, user_key):
    return fuzzy_match_entity(schema, user_key).value['name']


def fuzzy_match_entity(schema, user_key):
    ratios = list(
        map(lambda item: (
            item,
            1.0 if item.name.startswith(user_key) else SequenceMatcher(None, user_key,
                                                                       item.name).quick_ratio()),
            list(schema.EntityEnum)))
    ratios = sorted(ratios, key=lambda t: t[1])
    return ratios[-1][0]


def get_parent(file_or_folder):
    if hasattr(file_or_folder, 'parent_object_'):
        return file_or_folder.parent_object_
    return None


def select(context, target_type):
    return Select(context, target_type)


def get_entities(artifact):
    return {e['key']: e['value'] for e in artifact.entities}


def get_schema(model):
    current = model
    while current is not None:
        if isinstance(current, Dataset):
            return current._versioned_schema
        current = current.parent_object_
    return None


def get_sidecar_file(artifact, **entities):
    parent = artifact.get_parent()
    filters = dict(**artifact.get_entities())
    filters.update(**entities)
    return parent.query(**filters)


class PatchingSchemaPlugin(SchemaPlugin):
    def execute(self, schema):
        schema.Model.get_schema = get_schema
        schema.Model.__hash__ = lambda self: hash(tuple(self))
        schema.Folder.select = select
        schema.Folder.query = query
        schema.Folder.query_entities = query_entities
        schema.File.get_parent = get_parent
        schema.Folder.get_parent = get_parent
        schema.Artifact.has_entity = has_entity
        schema.Artifact.get_entity = get_entity
        schema.Artifact.get_entities = get_entities
        schema.Artifact.add_entity = lambda artifact, key, value: add_entity(schema, artifact, key, value)
        schema.Artifact.add_entities = lambda artifact, **kwargs: add_entities(schema, artifact, **kwargs)
        schema.Artifact.sidecar = get_sidecar_file
        schema.Folder.load_file_contents = load_file_contents
        schema.File.load_contents = load_contents
        schema.File.get_absolute_path = get_absolute_path_by_file
        schema.Folder.get_relative_path = _folder_get_relative_path
        schema.Folder.get_absolute_path = get_absolute_path
        schema.File.get_relative_path = _file_get_relative_path
        schema.Folder.remove_file = remove_file
        schema.Folder.create_artifact = create_artifact
        schema.Folder.create_folder = create_folder
        schema.Dataset.create_derivative = create_derivative
        schema.Folder.get_file = get_file
        schema.Folder.get_files = get_files
        schema.Folder.remove_folder = remove_folder
        schema.Folder.get_folder = get_folder
        schema.Folder.get_files_sorted = get_files_sorted
        schema.Folder.get_folders_sorted = get_folders_sorted
        schema.Model.to_generator = to_generator
        schema.Model.to_dict = to_dict
        schema.Model.iterancestors = iterancestors

        schema.get_model_classes = lambda: get_model_classes(schema)
        schema.get_members = lambda element_type, include_superclass=True: get_members(schema, element_type,
                                                                                       include_superclass)
        schema.process_entity_value = lambda key, value: process_entity_value(schema, key, value)
        schema.fuzzy_match_entity_key = lambda user_key: fuzzy_match_entity_key(schema, user_key)
        schema.fuzzy_match_entity = lambda user_key: fuzzy_match_entity(schema, user_key)

        schema.create_dataset = lambda base_dir=None, **kwargs: create_dataset(schema, base_dir=base_dir, **kwargs)



from ancpbids import load_dataset, _internal_validate_dataset
from ..base_test_case import *
from ancpbids.plugin import ValidationPlugin
from ancpbids.plugins import plugin_dsvalidator


class ValidationTestCase(BaseTestCase):
    def createSUT(self, ds_dir, rule_class):
        test_ds = load_dataset(ds_dir)
        # only test this plugin
        report = _internal_validate_dataset(test_ds, lambda plugin: isinstance(plugin, rule_class))
        self.assertTrue(isinstance(report, ValidationPlugin.ValidationReport))
        return report

    def test_validate_static_structure(self):
        report = self.createSUT(DS005_CONFLICT_DIR, plugin_dsvalidator.StaticStructureValidationPlugin)
        self.assertEqual(1, len(report.messages))
        self.assertTrue('dataset_description' in report.messages[0]['message'],
                        'dataset_description file should have been reported as missing')

    def test_validate_datatypes(self):
        report = self.createSUT(DS005_CONFLICT_DIR, plugin_dsvalidator.DatatypesValidationPlugin)
        self.assertEqual(2, len(report.messages))
        self.assertEqual("Unsupported datatype folder 'sub-01/abc'", report.messages[0]['message'].replace('\\', '/'))
        self.assertEqual("Unsupported datatype folder 'sub-01/xyz'", report.messages[1]['message'].replace('\\', '/'))

    def test_validation_entities(self):
        report = self.createSUT(RESOURCES_FOLDER + "/ds005_entities_validation",
                                plugin_dsvalidator.EntitiesValidationPlugin)
        self.assertEqual(2, len(report.messages))
        self.assertEqual(
            "Invalid entities order: expected=('sub', 'task', 'run'), found=('sub', 'run', 'task'), "
            "artifact=sub-01/func/sub-01_run-03_task-mixedgamblestask_events.tsv",
            report.messages[0]['message'].replace('\\', '/'))
        self.assertEqual("Invalid entity 'xyz' in artifact "
                         "'sub-01/func/sub-01_task-mixedgamblestask_run-03_xyz-001_events.tsv'",
                         report.messages[1]['message'].replace('\\', '/'))


if __name__ == '__main__':
    unittest.main()



import os.path
import warnings
from functools import partial
from typing import List, Union, Dict

import ancpbids
from ancpbids import ValidationPlugin
from . import load_dataset
from .query import query, query_entities
from .utils import resolve_segments, convert_to_relative

class BIDSLayout:
    """A convenience class to provide access to an in-memory representation of a BIDS dataset.

    .. code-block::

        dataset_path = 'path/to/your/dataset'
        layout = BIDSLayout(dataset_path)

    Parameters
    ----------
    ds_dir:
        the (absolute) path to the dataset to load
    """

    def __init__(self, ds_dir: str, **kwargs):
        self.dataset = load_dataset(ds_dir)
        self.schema = self.dataset.get_schema()

    def __getattr__(self, key):
        # replace arbitrary get functions with calls to get
        if key.startswith("get_"):
            return partial(self.get, "id", key[4:])

        # give up if the above don't work
        raise AttributeError(key)

    def get_metadata(self, path, include_entities=False, scope='all'):
        """Return metadata found in JSON sidecars for the specified file.

        Parameters
        ----------
        path : str
            Path to the file to get metadata for.
        include_entities : bool, optional
            If True, all available entities extracted
            from the filename (rather than JSON sidecars) are included in
            the returned metadata dictionary.
        scope : str or list, optional
            The scope of the search space. Each element must
            be one of 'all', 'raw', 'self', 'derivatives', or a
            BIDS-Derivatives pipeline name. Defaults to searching all
            available datasets.

        Returns
        -------
        dict
            A dictionary of key/value pairs extracted from all of the
            target file's associated JSON sidecars.

        Notes
        -----
        A dictionary containing metadata extracted from all matching .json
        files is returned. In cases where the same key is found in multiple
        files, the values in files closer to the input filename will take
        precedence, per the inheritance rules in the BIDS specification.

        """
        path = os.path.normpath(path)
        # make relative to dataset root, i.e., remove base path
        if path.startswith(self.dataset.base_dir_):
            path = path[len(self.dataset.base_dir_):].strip(os.sep)
        file = self.dataset.get_file(path)
        md = file.get_metadata(include_entities=include_entities)
        return md

    def get(self, return_type: str = 'object', target: str = None, scope: str = None,
            extension: Union[str, List[str]] = None, suffix: Union[str, List[str]] = None,
            **entities) -> Union[List[str], List[object]]:
        """Depending on the return_type value returns either paths to files that matched the filtering criteria
        or :class:`Artifact <ancpbids.model_v1_7_0.Artifact>` objects for further processing by the caller.

        Note that all provided filter criteria are AND combined, i.e. subj='02',task='lang' will match files containing
        '02' as a subject AND 'lang' as a task. If you provide a list of values for a criteria, they will be OR combined.

        .. code-block::

            file_paths = layout.get(subj='02', task='lang', suffix='bold', return_type='files')

            file_paths = layout.get(subj=['02', '03'], task='lang', return_type='files')

        Parameters
        ----------
        return_type:
            Either 'files' to return paths of matched files
            or 'object' to return :class:`Artifact <ancpbids.model_v1_7_0.Artifact>` object, defaults to 'object'

        target:
            Either `suffixes`, `extensions` or one of any valid BIDS entities key
            (see :class:`EntityEnum <ancpbids.model_v1_7_0.EntityEnum>`, defaults to `None`
        scope:
            a hint where to search for files
            If passed, only nodes/directories that match the specified scope will be
            searched. Possible values include:
            'all' (default): search all available directories.
            'derivatives': search all derivatives directories.
            'raw': search only BIDS-Raw directories.
            'self': search only the directly called BIDSLayout.
            <PipelineName>: the name of a BIDS-Derivatives pipeline.
        extension:
            criterion to match any files containing the provided extension only
        suffix:
            criterion to match any files containing the provided suffix only
        entities
            a list of key-values to match the entities of interest, example: subj='02',task='lang'

        Returns
        -------
            depending on the return_type value either paths to files that matched the filtering criteria
            or Artifact objects for further processing by the caller
        """
        folder = self.dataset
        return query(folder, return_type, target, scope, extension, suffix, **entities)

    def get_entities(self, scope: str = None, sort: bool = False) -> dict:
        """Returns a unique set of entities found within the dataset as a dict.
        Each key of the resulting dict contains a list of values (with at least one element).

        Example dict:
        .. code-block::

            {
                'sub': ['01', '02', '03'],
                'task': ['gamblestask']
            }

        Parameters
        ----------
        scope:
            see BIDSLayout.get()
        sort: default is `False`
            whether to sort the keys by name

        Returns
        -------
        dict
            a unique set of entities found within the dataset as a dict
        """
        return query_entities(self.dataset, scope, sort)

    def get_dataset_description(self, scope='self', all_=False) -> Union[List[Dict], Dict]:
        """Return contents of dataset_description.json.

        Parameters
        ----------
        scope : str
            The scope of the search space. Only descriptions of
            BIDSLayouts that match the specified scope will be returned.
            See :obj:`bids.layout.BIDSLayout.get` docstring for valid values.
            Defaults to 'self' --i.e., returns the dataset_description.json
            file for only the directly-called BIDSLayout.
        all_ : bool
            If True, returns a list containing descriptions for
            all matching layouts. If False (default), returns for only the
            first matching layout.

        Returns
        -------
        dict or list of dict
            a dictionary or list of dictionaries (depending on all_).
        """
        all_descriptions = self.dataset.select(self.schema.DatasetDescriptionFile).objects(as_list=True)
        if all_:
            return all_descriptions
        return all_descriptions[0] if all_descriptions else None

    def get_dataset(self) -> object:
        """
        Returns
        -------
            the in-memory representation of this layout/dataset
        """
        return self.dataset

    def write_derivative(self, derivative):
        """Writes the provided derivative folder to the dataset.
        Note that a 'derivatives' folder will be created if not present.

        Parameters
        ----------
        derivative:
            the derivative folder to write
        """
        assert isinstance(derivative, self.schema.DerivativeFolder)
        ancpbids.write_derivative(self.dataset, derivative)

    def validate(self) -> ValidationPlugin.ValidationReport:
        """Validates a dataset and returns a report object containing any detected validation errors.

        Example
        ----------

        .. code-block::

            report = layout.validate()
            for message in report.messages:
                print(message)
            if report.has_errors():
                raise "The dataset contains validation errors, cannot continue".

        Returns
        -------
            a report object containing any detected validation errors or warning
        """
        return ancpbids.validate_dataset(self.dataset)

    def get_files(self, scope='all'):
        """Get BIDSFiles for all layouts in the specified scope.

        Parameters
        ----------
        scope : str
            The scope of the search space. Indicates which
            BIDSLayouts' entities to extract.
            See :obj:`bids.layout.BIDSLayout.get` docstring for valid values.


        Returns:
            A dict, where keys are file paths and values
            are :obj:`bids.layout.BIDSFile` instances.

        """
        all_files = self.get(return_type="object", scope=scope)
        files = {file.get_absolute_path(): file for file in all_files}
        return files

    def get_file(self, filename, scope='all'):
        """Return the BIDSFile object with the specified path.

        Parameters
        ----------
        filename : str
            The path of the file to retrieve. Must be either an absolute path,
            or relative to the root of this BIDSLayout.
        scope : str or list, optional
            Scope of the search space. If passed, only BIDSLayouts that match
            the specified scope will be searched. See :obj:`BIDSLayout.get`
            docstring for valid values. Default is 'all'.

        Returns
        -------
        :obj:`bids.layout.BIDSFile` or None
            File found, or None if no match was found.
        """
        context = self.dataset
        filename = convert_to_relative(self.dataset, filename)
        if scope not in ['all', 'raw', 'self']:
            context, _ = resolve_segments(context, scope)
        return context.get_file(filename)




import inspect
import os

import ancpbids
from ancpbids.plugin import WritingPlugin, SchemaPlugin


class DatasetWritingPlugin(WritingPlugin):
    def execute(self, ds, target_dir: str, context_folder=None, src_dir: str = None):
        if context_folder is None and os.path.exists(target_dir) and len(os.listdir(target_dir)) > 0:
            raise ValueError("Directory not empty: " + target_dir)

        # set the target_dir as the base directory for creation
        if not hasattr(ds, 'base_dir_') or ds.base_dir_ is None:
            ds.base_dir_ = target_dir

        if context_folder is None:
            context_folder = ds
        if src_dir is None:
            src_dir = ds.get_absolute_path()

        self.schema = ds.get_schema()

        generator = context_folder.to_generator()
        for obj in generator:
            typ = type(obj)
            mapper_name = '_type_handler_' + typ.__name__
            if mapper_name not in _TYPE_MAPPERS:
                mapper_name = '_type_handler_default'
            mapper = _TYPE_MAPPERS[mapper_name]
            mapper(self, src_dir, target_dir, obj)
        # copy internal children (files/folders)
        self._type_handler_Folder(src_dir, target_dir, context_folder, traverse_children=True)

    def _type_handler_default(self, src_dir, target_dir, obj):
        if isinstance(obj, self.schema.Folder):
            self._type_handler_Folder(src_dir, target_dir, obj)
        elif isinstance(obj, self.schema.File):
            self._type_handler_File(src_dir, target_dir, obj)

    def _type_handler_File(self, src_dir, target_dir, file, new_file_name=None):
        abs_file_name = file.get_absolute_path()
        dir_name = os.path.dirname(abs_file_name)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        if hasattr(file, 'content') and callable(file.content):
            file.content(abs_file_name)
        else:
            ancpbids.utils.write_contents(abs_file_name, file)

    def _type_handler_Folder(self, src_dir, target_dir, folder, traverse_children=False):
        new_dir = os.path.join(target_dir, folder.get_relative_path())
        # the new directory may exist because model Artifacts/Folders are processed first
        if not os.path.exists(new_dir):
            os.makedirs(new_dir)

        if traverse_children:
            for child_folder in folder.folders:
                self._type_handler_Folder(src_dir, target_dir, child_folder)
            for child_file in folder.files:
                self._type_handler_File(src_dir, target_dir, child_file)

    def _get_ordered_entity_keys(self, artifact):
        schema = artifact.get_schema()
        entity_refs = artifact.entities

        schema_entities = list(map(lambda e: e.value['name'], list(schema.EntityEnum)))
        expected_key_order = {k: i for i, k in enumerate(schema_entities)}
        expected_order_key = {i: k for i, k in enumerate(schema_entities)}

        artifact_keys = list(map(lambda e: e.key, entity_refs))
        actual_keys_order = list(map(lambda k: expected_key_order[k], artifact_keys))
        expected = tuple(map(lambda k: expected_order_key[k], sorted(actual_keys_order)))
        return expected

    def _type_handler_Artifact(self, src_dir, target_dir, artifact):
        segments = []
        schema = artifact.get_schema()
        # add missing entities
        for ancestor in artifact.iterancestors():
            if isinstance(ancestor, schema.Folder):
                name = ancestor.name
                if name.startswith("ses-"):
                    artifact.add_entity('ses', name[4:])
                if name.startswith("sub-"):
                    artifact.add_entity('sub', name[4:])

        # sort according order defined in schema
        ordered_keys = self._get_ordered_entity_keys(artifact)
        for ok in ordered_keys:
            seg = '-'.join([ok, str(artifact.get_entity(ok))])
            segments.append(seg)
        segments.append(artifact.suffix)
        new_file_name = '_'.join(segments) + artifact.extension
        artifact.name = new_file_name
        self._type_handler_File(src_dir, target_dir, artifact, new_file_name)


_TYPE_MAPPERS = {name: obj for name, obj in inspect.getmembers(DatasetWritingPlugin) if
                 inspect.isfunction(obj) and obj.__name__.startswith('_type_handler_')}


def write_artifact(artifact):
    dummy_inst = DatasetWritingPlugin()
    dummy_inst._type_handler_Artifact(None, None, artifact)
    return artifact.get_absolute_path()


class ArtifactWritingSchemaPlugin(SchemaPlugin):
    def execute(self, schema):
        schema.Artifact.write = write_artifact



from ..base_test_case import *
import ancpbids


class PerformanceRegressionTestCase(BaseTestCase):
    def test_batch_loading(self):
        for i in range(0, 10):
            ds = ancpbids.load_dataset(DS005_DIR)



import fnmatch
import inspect
import os
import re

from .plugin_files_handlers import read_plain_text
from .. import utils
from ..plugin import DatasetPlugin, SchemaPlugin
from ..model_base import *


def lazy_contents_getter(lazy_loading_object):
    if callable(lazy_loading_object._contents):
        lazy_loading_object._contents = lazy_loading_object._contents()
    return lazy_loading_object._contents


def lazy_contents_setter(lazy_loading_object, value):
    lazy_loading_object._contents = value


def patch_contents_property(*types_to_patch):
    for type_to_patch in types_to_patch:
        # need a reference to the original contents property
        original_contents = type_to_patch.contents

        # replace with new property which uses the provided getter/setter
        type_to_patch.contents = property(lazy_contents_getter, lazy_contents_setter)

        # reverts the contents property to its original value, so, the next time contents is accessed, it will be reloaded
        # note that we must capture the outer variable original_contents as a default parameter value
        def unload(self, original=original_contents):
            self.contents = original

        # also allow to unload the contents
        type_to_patch.unload = unload


class PatchingSchemaPlugin(SchemaPlugin):
    def execute(self, schema):
        patch_contents_property(schema.MetadataFile, schema.MetadataArtifact, schema.TSVFile, schema.TSVArtifact,
                                schema.JsonFile)


class DatasetPopulationPlugin(DatasetPlugin):

    def execute(self, dataset, schema):
        base_dir = str(dataset.base_dir_)
        self.schema = schema
        self.options = dataset.options
        self._load_bidsignore(base_dir)

        # load file system structure
        self._load_folder(dataset, base_dir, base_dir)
        # transform files to artifacts, i.e. files containing entities in their name
        self._convert_files_to_artifacts(dataset)

        # handle special files
        self._handle_metadata_files(dataset)
        self._handle_tsv_files(dataset)

        # expand structure based on schema-files
        self._expand_members(dataset)
        # convert Folders within derivatives to DerivativeFolder
        self._convert_derivatives_folders(dataset.derivatives)

        # do optional stuff
        self._determine_artifact_datatype(dataset)

    def _determine_artifact_datatype(self, dataset):
        if not self.options.infer_artifact_datatype:
            return
        datatype_folders = dataset.select(DatatypeFolder).objects()
        for folder in datatype_folders:
            artifacts = folder.select(Artifact).objects()
            for artifact in artifacts:
                artifact.datatype = folder.name

    def _load_bidsignore(self, base_dir):
        self.bidsignore = lambda relative_path: False
        if self.options.ignore:
            patterns = []
            if isinstance(self.options.ignore, bool):
                bidsignore_file = os.path.join(base_dir, ".bidsignore")
                if os.path.exists(bidsignore_file):
                    patterns = read_plain_text(bidsignore_file)
            elif isinstance(self.options.ignore, list):
                patterns = self.options.ignore

            if patterns:
                patterns = list(map(lambda pattern: pattern.strip(), patterns))
                # TODO cleanup invalid filter such as empty lines or comments
                self.bidsignore = lambda relative_path: next(
                    filter(lambda pattern: fnmatch.fnmatch(relative_path, pattern), patterns), False)

    def _handle_metadata_files(self, folder):
        if not isinstance(folder, Folder):
            return
        for file in filter(lambda f: f.name.endswith(".json"), folder.files):
            if isinstance(file, Artifact):
                mdfile = MetadataArtifact()
            else:
                mdfile = MetadataFile()

            mdfile.parent_object_ = folder
            mdfile.update(file)

            # Load contents based on lazy loading setting
            if self.options.lazy_loading:
                mdfile.contents = lambda _mdfile=mdfile: _mdfile.load_contents()
            else:
                mdfile.contents = mdfile.load_contents()

            folder.files.remove(file)
            folder.files.append(mdfile)

        for child in folder.folders:
            self._handle_metadata_files(child)

    def _handle_tsv_files(self, folder):
        if not isinstance(folder, Folder):
            return
        for file in filter(lambda f: f.name.endswith(".tsv"), folder.files):
            if isinstance(file, Artifact):
                newfile = TSVArtifact()
            else:
                newfile = TSVFile()

            newfile.parent_object_ = folder
            newfile.update(file)

            if self.options.lazy_loading:
                newfile.contents = lambda _newfile=newfile: _newfile.load_contents()
            else:
                newfile.contents = newfile.load_contents()

            folder.files.remove(file)
            folder.files.append(newfile)

        for child in folder.folders:
            self._handle_tsv_files(child)

    def _type_handler_JsonFile(self, parent, member, allow_lazy=True):
        name = member['name']
        file_name = name + '.json'
        file = parent.get_file(file_name)
        if not file:
            return

        # create a placeholder json file within the parent folder and remove the File object
        model_type = member['type']
        json_file = model_type()
        setattr(parent, member['name'], json_file)
        parent.remove_file(file_name)
        json_file.parent_object_ = parent
        json_file.name = file_name

        # if lazy-loading=True, defer mapping/constructing the json object
        if allow_lazy and self.options.lazy_loading:
            json_file.contents = lambda _jf=json_file, _f=file, _mt=model_type: self._contents_loader(_jf, _f, _mt)
        else:
            json_file.contents = self._contents_loader(json_file, file, model_type)

    def _contents_loader(self, json_file, original_file, model_type):
        json_object = original_file.load_contents()
        self._map_object(model_type, json_object, json_file)
        return json_object

    def _convert_derivatives_folders(self, parent):
        if not parent:
            return
        for i, folder in enumerate(list(parent.folders)):
            dfolder = DerivativeFolder()
            dfolder.parent_object_ = parent
            dfolder.update(folder)
            parent.folders[i] = dfolder
            self._convert_derivatives_folders(dfolder)
            self._expand_members(dfolder)

    def _convert_files_to_artifacts(self, parent):
        for i, file in enumerate(parent.files):
            artifact = self._convert_to_artifact(file)
            if not artifact:
                continue
            artifact.parent_object_ = parent
            parent.files[i] = artifact
        for folder in parent.folders:
            self._convert_files_to_artifacts(folder)

    def _convert_to_artifact(self, file):
        parts = utils.parse_bids_name(file.name)
        if not parts:
            return None
        artifact = Artifact()
        artifact.name = file.name
        for key, value in parts['entities'].items():
            entity = EntityRef()
            entity.key = key
            value = self.schema.process_entity_value(key, value)
            entity.value = value
            artifact.entities.append(entity)
        artifact.suffix = parts['suffix']
        artifact.extension = parts['extension']
        return artifact

    def _handle_direct_folders(self, parent, member, pattern, new_type):
        if not isinstance(parent, Folder):
            return
        parent_folders = parent.get_folders_sorted()
        folders = list(filter(lambda f: re.match(pattern, f.name), parent_folders))
        for folder in folders:
            obj = new_type()
            obj.name = folder.name
            obj.files = folder.files
            for ofile in obj.files:
                ofile.parent_object_ = obj
            obj.folders = folder.folders
            for ofolder in obj.folders:
                ofolder.parent_object_ = obj
            parent.remove_folder(folder.name)
            obj.parent_object_ = parent
            if member['max'] > 1:
                getattr(parent, member['name']).append(obj)
            else:
                setattr(parent, member['name'], obj)
            self._expand_members(obj)

    def _expand_member(self, parent, member):
        typ = member['type']
        if not issubclass(typ, Model):
            return
        mapper_name = '_type_handler_%s' % typ.__name__
        if mapper_name not in _TYPE_MAPPERS:
            mapper_name = '_type_handler_default'
        mapper = _TYPE_MAPPERS[mapper_name]
        # do not allow lazy-loading for member fields as those are part of the BIDS schema
        mapper(self, parent, member, allow_lazy=False)

    def _expand_members(self, folder):
        members = self.schema.get_members(type(folder))
        for member in members:
            self._expand_member(folder, member)

    def _load_folder(self, parent, dir_path, ds_path):
        for root, directories, files in os.walk(dir_path):
            rel_base = root[len(ds_path):]
            for directory in sorted(directories):
                directory_ds_rel_path = '/'.join([rel_base, directory])[1:]
                if self.bidsignore(directory_ds_rel_path):
                    continue
                folder = Folder()
                folder.parent_object_ = parent
                folder.name = directory
                parent.folders.append(folder)
                self._load_folder(folder, '/'.join([root, directory]), ds_path)
            for file in sorted(files):
                file_ds_rel_path = '/'.join([rel_base, file])[1:]
                if self.bidsignore(file_ds_rel_path):
                    continue
                model_file = File()
                model_file.parent_object_ = parent
                model_file.name = file
                parent.files.append(model_file)
            # do not traverse into sub-dirs as they have been already processed recursively
            break

    def _type_handler_default(self, parent, member, allow_lazy=True):
        typ = member['type']
        if issubclass(typ, JsonFile):
            self._type_handler_JsonFile(parent, member, allow_lazy)
        elif issubclass(typ, Folder):
            pattern = '.*'
            meta = member['meta']
            if 'name_pattern' in meta:
                pattern = meta['name_pattern']
            self._handle_direct_folders(parent, member, pattern=pattern, new_type=typ)

    def _type_handler_File(self, parent, member, allow_lazy=True):
        if not isinstance(parent, Folder):
            return
        file_name = member['name']
        meta = member['meta']
        if 'name_pattern' in meta:
            file_name = meta['name_pattern']
        file = parent.get_file(file_name)
        if file:
            setattr(parent, member['name'], file)
            parent.remove_file(file.name)

    def _type_handler_Artifact(self, parent, member, allow_lazy=True):
        if not isinstance(parent, Folder):
            return
        attr = getattr(parent, member['name'])
        multi = isinstance(attr, list)
        name = member['name']
        files = parent.files if multi else list(filter(lambda f: f.name == name, parent.files))
        for file in files:
            if not isinstance(file, Artifact):
                continue
            file.parent_object_ = parent
            parent.remove_file(file.name)
            if multi:
                attr.append(file)
            else:
                setattr(parent, member['name'], file)

    def _type_handler_Folder(self, parent, member, allow_lazy=True):
        if not isinstance(parent, Folder):
            return
        name = member['name']
        folder = parent.get_folder(name)
        if folder:
            setattr(parent, name, folder)
            parent.remove_folder(name)

    def _map_object(self, model_type, json_object, target=None):
        if target is None:
            target = model_type()
        members = self.schema.get_members(model_type, True)
        actual_props = json_object.keys()
        direct_props = list(map(lambda m: (m['name'], m), members))
        for prop_name, prop in direct_props:
            if prop_name in actual_props:
                value_type = prop['type']
                value = json_object[prop_name]
                if isinstance(value, list) and len(value) > 0:
                    value = list(map(lambda o: self._map_object(value_type, o) if isinstance(o, dict) else o, value))
                if isinstance(value, dict):
                    value = self._map_object(value_type, value)
                setattr(target, prop_name, value)
        return target


_TYPE_MAPPERS = {name: obj for name, obj in inspect.getmembers(DatasetPopulationPlugin) if
                 inspect.isfunction(obj) and obj.__name__.startswith('_type_handler_')}



import unittest

from ancpbids import model_v1_8_0, model_latest, load_dataset, load_schema
from ..base_test_case import BaseTestCase, DS005_DIR, DS005_SMALL_DIR


class SchemaTestCase(BaseTestCase):
    def test_entity_matching(self):
        self.assertEqual('sub', model_latest.fuzzy_match_entity_key('sub'))
        self.assertEqual('sub', model_latest.fuzzy_match_entity_key('subject'))
        self.assertEqual('sub', model_latest.fuzzy_match_entity_key('subjects'))
        self.assertEqual('sub', model_latest.fuzzy_match_entity_key('subjs'))

        self.assertEqual('desc', model_latest.fuzzy_match_entity_key('des'))
        self.assertEqual('desc', model_latest.fuzzy_match_entity_key('dscr'))
        self.assertEqual('desc', model_latest.fuzzy_match_entity_key('descriptions'))

    def test_schema_versions(self):
        ds_latest = load_dataset(DS005_DIR)
        schema_latest = ds_latest.get_schema()
        self.assertEqual(schema_latest, model_latest)
        self.assertEqual('1.10.0', schema_latest.VERSION)

        ds_old = load_dataset(DS005_SMALL_DIR)
        schema_old = ds_old.get_schema()
        self.assertEqual(schema_old, model_v1_8_0)
        self.assertEqual('v1.8.0', schema_old.VERSION)

    def test_load_schema(self):
        schema_latest = load_schema(DS005_DIR)
        self.assertEqual(schema_latest, model_latest)
        self.assertEqual('1.10.0', schema_latest.VERSION)

        schema_v180 = load_schema(DS005_SMALL_DIR)
        self.assertEqual(schema_v180, model_v1_8_0)
        self.assertEqual('v1.8.0', schema_v180.VERSION)

        # The classes of each schema are separate identities
        # assert on arbitrary class
        self.assertFalse(schema_latest.DatatypeEnum == schema_v180.DatatypeEnum)

    def test_v190_motion_modality_exists(self):
        from ancpbids import model_v1_9_0
        # in 1.9.0 the motion modality was added
        self.assertTrue("motion" in [e.name for e in model_v1_9_0.ModalityEnum])

    def test_v110_mrs_modality_exists(self):
        from ancpbids import model_v1_10_0
        # in 1.10.0 the mrs modality was added
        self.assertTrue("motion" in [e.name for e in model_v1_10_0.ModalityEnum])



if __name__ == '__main__':
    unittest.main()



import os
import re
import sys
from collections import OrderedDict
from fnmatch import fnmatch
from typing import Union, List

from ancpbids.utils import resolve_segments
from ancpbids.model_base import *


class Expr:
    def convert_value(self, value):
        if hasattr(self, 'value_converter'):
            value = self.value_converter(value)
        return value


class BoolExpr(Expr):
    pass


class CompExpr(Expr):
    pass


class TrueExpr(BoolExpr):
    def __init__(self, *args):
        pass

    def eval(self, context) -> bool:
        return True


class AnyExpr(BoolExpr):
    def __init__(self, *bool_ops: CompExpr):
        self.bool_ops = bool_ops

    def eval(self, context) -> bool:
        return any([op.eval(context) for op in self.bool_ops])


class AllExpr(BoolExpr):
    def __init__(self, *bool_ops: CompExpr):
        self.bool_ops = bool_ops

    def eval(self, context) -> bool:
        for op in self.bool_ops:
            if not op.eval(context):
                # early exit to prevent evaluating remaining ops
                return False
        return True


class EqExpr(CompExpr):
    def __init__(self, attr: property, value):
        self.attr = attr
        self.value = value

    def eval(self, context) -> bool:
        value = self.attr.fget(context)
        value = self.convert_value(value)
        return self.value == value


class ReExpr(CompExpr):
    def __init__(self, attr: property, regex_pattern: str):
        self.attr = attr
        self.regex_pattern = re.compile(regex_pattern)

    def eval(self, context) -> bool:
        value = self.attr.fget(context)
        value = self.convert_value(value)
        value = str(value)
        return self.regex_pattern.search(value)


class CustomOpExpr(CompExpr):
    def __init__(self, op):
        self.op = op

    def eval(self, context) -> bool:
        return self.op(context)


class FnMatchExpr(CompExpr):
    def __init__(self, attr: property, pattern):
        self.attr = attr
        self.pattern = pattern

    def eval(self, context) -> bool:
        value = self.attr.fget(context)
        value = self.convert_value(value)
        value = str(value)
        return value is not None and fnmatch(value, self.pattern)


class EntityExpr(CompExpr):
    def __init__(self, schema, key, pattern, op=FnMatchExpr):
        self.schema = schema
        if pattern is not None:
            pattern = schema.process_entity_value(key, pattern)
            if isinstance(pattern, list):
                pattern = list(map(lambda v: str(v), pattern))
            else:
                pattern = str(pattern)
        self.key = key
        self.pattern = pattern
        self.op = op(schema.EntityRef.value, pattern)
        self.op.value_converter = lambda v: schema.process_entity_value(key, v)

    def eval(self, context) -> bool:
        if not isinstance(context, self.schema.Artifact):
            # for non-Artifacts, for example File, just return false
            return False
        ents = list(filter(lambda e: e.key == self.key.value['name'], context.entities))
        if not ents:
            # the entity must not exist
            if self.pattern is None:
                return True
            return False
        if self.pattern is None:
            # the entity must not exist, but we found one
            return False
        target_key = ents[0]
        return self.op.eval(target_key)


class Select:
    def __init__(self, context, filter_type):
        self.schema = context.get_schema()
        self.context = context
        self.filter_type = filter_type
        self._where = TrueExpr()
        self._subtree = TrueExpr()

    def subtree(self, bool_expr: BoolExpr):
        self._subtree = bool_expr

    def where(self, bool_expr: BoolExpr):
        self._where = bool_expr
        return self

    def _exec(self, callback, depth=sys.maxsize):
        for m in self.context.to_generator(filter_=lambda o: self._subtree.eval(o), depth=depth):
            if isinstance(m, self.filter_type) and self._where.eval(m):
                yield callback(m)

    def get_file_paths(self, depth=sys.maxsize):
        return self._exec(self.schema.File.get_relative_path, depth=depth)

    def get_file_paths_absolute(self, depth=sys.maxsize):
        return self._exec(self.schema.File.get_absolute_path, depth=depth)

    def get_artifacts(self):
        # TODO filter by Artifact instances
        return self.objects()

    def objects(self, as_list=False, depth=sys.maxsize):
        result = self._exec(callback=lambda m: m, depth=depth)
        if as_list:
            result = list(result)
        return result


def _to_any_expr(value, ctor, converter=lambda v: v):
    # if the value is a list, then wrap it in an AnyExpr
    if isinstance(value, list):
        ops = []
        for v in value:
            ops.append(ctor(converter(v)))
        return AnyExpr(*ops)
    # else just return using the constructor function
    return ctor(converter(value))


def _require_artifact(schema, expr) -> AllExpr:
    """Wraps the provided expression in an expression that makes sure the context of evaluation is an Artifact.

    Parameters
    ----------
    expr :
        the expression to wrap

    Returns
    -------
        a wrapping expression to make sure that the provided object is an instance of Artifact
    """
    return AllExpr(CustomOpExpr(lambda m: isinstance(m, schema.Artifact)), expr)


def query(folder, return_type: str = 'object', target: str = None, scope: str = None,
          extension: Union[str, List[str]] = None, suffix: Union[str, List[str]] = None,
          regex_search=False, sorter=None,
          **entities) -> Union[List[str], List[object]]:
    """Depending on the return_type value returns either paths to files that matched the filtering criteria
    or :class:`Artifact <ancpbids.model_v1_7_0.Artifact>` objects for further processing by the caller.

    Note that all provided filter criteria are AND combined, i.e. subj='02',task='lang' will match files containing
    '02' as a subject AND 'lang' as a task. If you provide a list of values for a criteria, they will be OR combined.

    .. code-block::

        file_paths = layout.get(subj='02', task='lang', suffix='bold', return_type='files')

        file_paths = layout.get(subj=['02', '03'], task='lang', return_type='files')

    Parameters
    ----------
    folder:
            an entry-point of type Folder to search within
    return_type:
        Either 'files' to return paths of matched files
        or 'object' to return :class:`Artifact <ancpbids.model_v1_7_0.Artifact>` object, defaults to 'object'

    target:
        Either `suffixes`, `extensions` or one of any valid BIDS entities key
        (see :class:`EntityEnum <ancpbids.model_v1_7_0.EntityEnum>`, defaults to `None`
    scope:
        a hint where to search for files
        If passed, only nodes/directories that match the specified scope will be
        searched. Possible values include:
        'all' (default): search all available directories.
        'derivatives': search all derivatives directories.
        'raw': search only BIDS-Raw directories.
        'self': search only the directly called BIDSLayout.
        <PipelineName>: the name of a BIDS-Derivatives pipeline.
    extension:
        criterion to match any files containing the provided extension only
    suffix:
        criterion to match any files containing the provided suffix only
    entities
        a list of key-values to match the entities of interest, example: subj='02',task='lang'

    Returns
    -------
        depending on the return_type value either paths to files that matched the filtering criteria
        or Artifact objects for further processing by the caller
    """
    if scope is None:
        scope = 'all'
    if return_type == 'id':
        if not target:
            raise ValueError("return_type=id requires the target parameter to be set")

    schema = folder.get_schema()
    context = folder
    ops = []
    target_type = File
    if scope not in ['all', 'raw', 'self']:
        context, _ = resolve_segments(folder, scope, False)

    if not context:
        return None

    select = context.select(target_type)

    if scope == 'raw':
        # the raw scope does not search in derivatives folder but everything else
        select.subtree(CustomOpExpr(lambda m: not isinstance(m, DerivativeFolder)))

    result_extractor = None
    if target:
        if target in 'suffixes':
            suffix = '*'
            result_extractor = lambda artifacts: [a.suffix for a in artifacts if a.suffix]
        elif target in 'extensions':
            extension = '*'
            result_extractor = lambda artifacts: [a.extension for a in artifacts if a.extension]
        else:
            target = schema.fuzzy_match_entity_key(target)
            entities = {**entities, target: '*'}
            result_extractor = lambda artifacts: [entity.value for a in artifacts for entity in
                                                  filter(lambda e: e.key == target, a.entities) if a.entities]

    search_operator = FnMatchExpr
    if regex_search:
        search_operator = ReExpr

    for k, v in entities.items():
        entity_key = schema.fuzzy_match_entity(k)
        v = schema.process_entity_value(entity_key, v)
        ops.append(
            _require_artifact(schema,
                              _to_any_expr(v, lambda val: EntityExpr(schema, entity_key, val, op=search_operator))))

    if extension:
        converter = lambda v: "." + v if v != "*" and not v.startswith(".") else v
        any_expr = _to_any_expr(extension, lambda ext: search_operator(Artifact.extension, ext), converter)
        require_expr = _require_artifact(schema, any_expr)
        ops.append(require_expr)

    if suffix:
        ops.append(
            _require_artifact(schema,
                              _to_any_expr(suffix, lambda suf: search_operator(Artifact.suffix, suf))))

    select.where(AllExpr(*ops))

    search_depth = sys.maxsize
    if scope == "self":
        search_depth = 1

    if return_type:
        if return_type.startswith("file"):
            return list(select.get_file_paths_absolute())
        elif return_type == 'dir':
            result = filter(lambda o: isinstance(o, File), select.objects(depth=search_depth))
            return set(map(lambda a: a.get_parent().get_relative_path(), result))

    artifacts = select.objects(depth=search_depth)
    if sorter is None:
        sorter = lambda artifact: artifact.name
    if callable(sorter):
        artifacts = sorted(artifacts, key=sorter)
    if result_extractor:
        return sorted(set(result_extractor(artifacts)))
    return list(artifacts)


def query_entities(folder, scope: str = None, sort: bool = False, long_form=True) -> dict:
    """Returns a unique set of entities found within the dataset as a dict.
    Each key of the resulting dict contains a list of values (with at least one element).

    Example dict:
    .. code-block::

        {
            'sub': ['01', '02', '03'],
            'task': ['gamblestask']
        }

    Parameters
    ----------
    folder:
        an entry-point of type Folder to search within
    scope:
        see BIDSLayout.get()
    sort: default is `False`
        whether to sort the keys by name

    Returns
    -------
    dict
        a unique set of entities found within the dataset as a dict
    """
    schema = folder.get_schema()
    artifacts = filter(lambda m: isinstance(m, schema.Artifact), query(folder, scope=scope))
    result = {}
    for e in [e for a in artifacts for e in a.entities]:
        key = e.key
        if key not in result:
            result[key] = set()
        result[key].add(e.value)
    if long_form:
        known_entities = {e.value['name']: e.name for e in list(schema.EntityEnum)}
        result = {known_entities[k] if k in known_entities else k: v for k, v in result.items()}
    if sort:
        result = {k: sorted(v) for k, v in sorted(result.items())}
    return result



import logging
import os
import sys
from dataclasses import dataclass
from typing import Union, List, Optional

from ancpbids import plugins
from ancpbids import utils
from . import model_v1_8_0
from . import model_v1_9_0
from . import model_v1_10_0
# latest stable supported BIDS schema
from ancpbids import model_v1_10_0 as model_latest

from .plugin import get_plugins, load_plugins_by_package, DatasetPlugin, WritingPlugin, ValidationPlugin, SchemaPlugin, \
    FileHandlerPlugin
from .query import BoolExpr, Select, EqExpr, AnyExpr, AllExpr, ReExpr, CustomOpExpr, \
    EntityExpr

LOGGER = logging.getLogger("ancpbids")


# ENTITIES_PATTERN = regex.compile(r'(([^\W_]+)-([^\W_]+)_)+([^\W_]+)((\.[^\W_]+)+)')

@dataclass
class DatasetOptions(dict):
    """All options that can be set to influence handling of reading/writing a dataset from/to file system."""
    infer_artifact_datatype: bool = False
    """If True, will determine the datatype an Artifact is contained in, either directly or within a sub-directory.
        For example, given the path "sub-02/func/sub-02_task-mixedgamblestask_run-01_bold.nii.gz", the datatype will be "func".
        
        By default, this option is set to False as it may have a negative performance impact.
    """

    ignore: Union[bool, List[str]] = False
    """If a .bidsignore file is available at the root, all resources (files/folders) matching the filters
        in that file will not be added to the in-memory graph. Alternatively, a list of fnmatch patterns can be provided.
        
        By default, this option is set to False as it may have a negative performance impact."""

    lazy_loading: bool = False
    """If True, file contents will be loaded only when accessed, not during dataset initialization.
        This can significantly improve memory usage and loading performance for large datasets.

        By default, this option is set to False to maintain backward compatibility."""


def load_dataset(base_dir: str, options: Optional[DatasetOptions] = None):
    """Loads a dataset given its directory path on the file system.

    .. code-block::

        from ancpbids import load_dataset, validate_dataset
        dataset_path = 'path/to/your/dataset'
        dataset = load_dataset(dataset_path, DatasetOptions(ignore=False, infer_artifact_datatype=True))

    Parameters
    ----------
    base_dir:
        the dataset path to load from
    options:
        the options to use, see :py:class:`DatasetOptions` class for available options

    Returns
    -------
    str
        a Dataset object depending on the used schema which represents the dataset as an in-memory graph
    """
    if not os.path.isdir(base_dir):
        raise ValueError("Invalid Directory")
    schema = load_schema(base_dir)
    ds = schema.Dataset()
    ds._versioned_schema = schema
    ds.options = options
    if ds.options is None:
        ds.options = DatasetOptions()
    ds.name = os.path.basename(base_dir)
    ds.base_dir_ = base_dir
    dataset_plugins = get_plugins(DatasetPlugin)
    for dsplugin in dataset_plugins:
        dsplugin.execute(ds, schema)
    return ds


def load_schema(base_dir):
    """Loads a BIDS schema object which represents the static/formal definition of the BIDS specification.

    As per BIDS spec, a BIDS compliant dataset must have a BIDSVersion field in the dataset_description.json
    file at the top level. This field is used to determine which BIDS schema to load.

    In case the BIDSVersion field is missing or not supported, the earliest supported schema will be returned.

    Parameters
    ----------
    base_dir:
        The dataset directory path where a dataset_description.json must be located.

    Returns
    -------
    object
        A BIDS schema object which represents the static/formal definition of the BIDS specification.
    """
    ds_descr_path = os.path.join(base_dir, "dataset_description.json")
    if os.path.exists(ds_descr_path):
        ds_descr = utils.load_contents(ds_descr_path)
        if isinstance(ds_descr, dict) and 'BIDSVersion' in ds_descr:
            schema_version = ds_descr['BIDSVersion']
            schema_version = schema_version.replace('.', '_')
            schema_name = f'ancpbids.model_v{schema_version}'
            if schema_name in sys.modules:
                schema = sys.modules[schema_name]
                return schema
    # assume using the latest supported schema
    return model_latest


def save_dataset(ds: object, target_dir: str, context_folder=None):
    """Copies the dataset graph into the provided target directory.

    EXPERIMENTAL/UNSTABLE

    Parameters
    ----------
    ds:
        the dataset graph to save
    target_dir
        the target directory to save to
    context_folder
        a folder node within the dataset graph to limit to

    """
    dataset_plugins = get_plugins(WritingPlugin)
    for dsplugin in dataset_plugins:
        dsplugin.execute(ds, target_dir, context_folder=context_folder)


def validate_dataset(dataset) -> ValidationPlugin.ValidationReport:
    """Validates a dataset and returns a report object containing any detected validation errors.

    Example:

    .. code-block::

        report = validate_dataset(dataset)
        for message in report.messages:
            print(message)
        if report.has_errors():
            raise "The dataset contains validation errors, cannot continue".

    Parameters
    ----------
    dataset:
        the dataset to validate

    Returns
    -------
    ValidationPlugin.ValidationReport
        a report object containing any detected validation errors or warning
    """
    return _internal_validate_dataset(dataset)


def _internal_validate_dataset(dataset, plugin_acceptor=None):
    validation_plugins = get_plugins(ValidationPlugin)
    report = ValidationPlugin.ValidationReport()
    for validation_plugin in validation_plugins:
        # if plugin is disabled, skip it
        if callable(plugin_acceptor) and not plugin_acceptor(validation_plugin):
            continue
        validation_plugin.execute(dataset=dataset, report=report)
    return report


def write_derivative(ds, derivative):
    """Writes the provided derivative folder to the dataset.
    Note that a 'derivatives' folder will be created if not present.

    Parameters
    ----------
    ds:
        the dataset object to extend
    derivative:
        the derivative folder to write
    """
    save_dataset(ds, target_dir=ds.get_absolute_path(), context_folder=derivative)


# load system plugins using lowest rank value
load_plugins_by_package(plugins, ranking=0, system=True)

# execute all SchemaPlugins, these plugins may monkey-patch the schema
for pl in get_plugins(SchemaPlugin):
    for schema in [model_v1_8_0, model_v1_9_0, model_v1_10_0]:
        pl.execute(schema)

# load file handler plugins
for pl in get_plugins(FileHandlerPlugin):
    pl.execute(utils.FILE_READERS, utils.FILE_WRITERS)

from .pybids_compat import BIDSLayout

select = Select
any_of = AnyExpr
all_of = AllExpr
eq = EqExpr
re = ReExpr
op = CustomOpExpr
entity = EntityExpr

from . import _version

__version__ = _version.get_versions()['version']



import os
import unittest

TEST_FOLDER = os.path.dirname(__file__)
RESOURCES_FOLDER = TEST_FOLDER + "/data"

DS005_DIR = RESOURCES_FOLDER + "/ds005"
DS005_CONFLICT_DIR = RESOURCES_FOLDER + "/ds005_conflict"
DS005_SMALL_DIR = RESOURCES_FOLDER + "/ds005-small"
DS005_SMALL2_DIR = RESOURCES_FOLDER + "/ds005-small2"
SYNTHETIC_DIR = RESOURCES_FOLDER + "/synthetic"
ENTITIES_DIR = RESOURCES_FOLDER + "/ds005_entities_validation"
DS005_DIR_IGNORED_RESOURCES = RESOURCES_FOLDER + "/ds005_ignore"
DS7t_trt_DIR = RESOURCES_FOLDER + "/7t_trt"


class BaseTestCase(unittest.TestCase):
    pass



import os.path
import unittest
import tempfile

from ancpbids import utils, load_dataset
from ..base_test_case import BaseTestCase


class FetchDatasetTestCase(BaseTestCase):
    def test_fetch_dataset(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            ds_path = utils.fetch_dataset('ds003483', output_dir=temp_dir)
            self.assertEqual(os.path.join(temp_dir, 'ds003483'), ds_path)
            dataset = load_dataset(ds_path)

            # some basic checks to make sure the dataset is downloaded and unzipped as expected
            self.assertEqual('ds003483', dataset.name)

            entities = dataset.query_entities()
            self.assertEqual(21, len(entities["subject"]))
            suffixes = dataset.query(target="suffixes")
            self.assertEqual(['channels', 'coordsystem', 'events', 'meg', 'scans'], suffixes)


if __name__ == '__main__':
    unittest.main()



from ancpbids.plugin import ValidationPlugin


class StaticStructureValidationPlugin(ValidationPlugin):
    def execute(self, dataset, report: ValidationPlugin.ValidationReport):
        self.schema = dataset.get_schema()
        gen = dataset.to_generator()
        for obj in gen:
            top_path = obj.get_relative_path().replace("\\", "/") if isinstance(obj, (
                self.schema.File, self.schema.Folder)) else '???'
            members = self.schema.get_members(type(obj))
            for member in members:
                typ = member['type']
                name = member['name']
                lb = member['min']
                ub = member['max']
                val = getattr(obj, name)
                use = member['use']
                if (lb > 0 or use == 'required') and not val:
                    report.error(f"Missing required node {name} at {top_path}.",
                                 obj)
                if use == 'recommended' and not val:
                    report.warn(f"Missing recommended field {name} at {top_path}.",
                                obj)


class DatatypesValidationPlugin(ValidationPlugin):
    def execute(self, dataset, report: ValidationPlugin.ValidationReport):
        invalid = []
        valid_datatypes = [v.name for v in dataset.get_schema().DatatypeEnum.__members__.values()]
        for subject in dataset.subjects:
            invalid.extend([f for f in subject.datatypes if f.name not in valid_datatypes])
            for session in subject.sessions:
                invalid.extend([f for f in session.datatypes if f.name not in valid_datatypes])

        for folder in invalid:
            report.error("Unsupported datatype folder '%s'" % folder.get_relative_path(),
                         folder)


class EntitiesValidationPlugin(ValidationPlugin):
    def execute(self, dataset, report: ValidationPlugin.ValidationReport):
        schema = dataset.get_schema()
        artifacts = dataset.select(schema.Artifact).get_artifacts()
        entities = list(map(lambda e: e.value['name'], list(schema.EntityEnum)))
        expected_key_order = {k: i for i, k in enumerate(entities)}
        expected_order_key = {i: k for i, k in enumerate(entities)}
        for artifact in artifacts:
            entity_refs = artifact.entities
            found_invalid_key = False
            for ref in entity_refs:
                if ref.key not in entities:
                    report.error(
                        "Invalid entity '%s' in artifact '%s'" % (ref.key, artifact.get_relative_path()), artifact)
                    found_invalid_key = True
            if found_invalid_key:
                # we cannot check the order of entities if invalid entity found
                continue
            # now, check if order of entities matches order in schema
            keys = list(map(lambda e: e.key, entity_refs))
            actual_keys_order = list(map(lambda k: expected_key_order[k], keys))
            for i in range(0, len(actual_keys_order) - 1):
                if actual_keys_order[i] > actual_keys_order[i + 1]:
                    expected = tuple(map(lambda k: expected_order_key[k], sorted(actual_keys_order)))
                    report.error(
                        "Invalid entities order: expected=%s, found=%s, artifact=%s" % (
                            expected, tuple(keys), artifact.get_relative_path()), artifact)
                    break


class SuffixesValidationPlugin(ValidationPlugin):
    def execute(self, dataset, report: ValidationPlugin.ValidationReport):
        pass



import os.path

import ancpbids
from ancpbids import select, re, any_of, all_of, eq, op, entity
from ..base_test_case import *


class QueryTestCase(BaseTestCase):
    def test_entities_formatting(self):
        layout = ancpbids.load_dataset(ENTITIES_DIR)
        files = layout.query(sub='02', run='3', return_type='filename')
        self.assertEqual(1, len(files))
        self.assertTrue(files[0].endswith("sub-02_task-abc_run-00003_events.tsv"))

        files = layout.query(sub='02', run=['000000003'], return_type='filename')
        self.assertEqual(1, len(files))
        self.assertTrue(files[0].endswith("sub-02_task-abc_run-00003_events.tsv"))

        # should also handle invalid formats, i.e. run not as an index but a label
        files = layout.query(sub='02', run='xyz', return_type='filename')
        self.assertEqual(1, len(files))
        self.assertTrue(files[0].endswith("sub-02_task-abc_run-xyz_events.tsv"))

    def test_bidslayout_entities_any(self):
        layout = ancpbids.load_dataset(ENTITIES_DIR)
        files = layout.query(sub='*', suffix='test', task='abc', return_type='filename')
        self.assertEqual(2, len(files))
        self.assertTrue(files[0].endswith("sub-bar_task-abc_test.txt"))
        self.assertTrue(files[1].endswith("sub-foo_task-abc_test.txt"))

    def test_bidslayout_subjects_filtered(self):
        layout = ancpbids.load_dataset(ENTITIES_DIR)
        subjects = layout.query(target="sub", task='abc')
        self.assertEqual(3, len(subjects))
        self.assertListEqual(['02', 'bar', 'foo'], subjects)

    def test_bidslayout(self):
        layout = ancpbids.load_dataset(DS005_DIR)
        ents = layout.query_entities()

        subjects = ents["subject"]
        subjects_expected = {'%02d' % i for i in range(1, 17)}
        self.assertSetEqual(subjects_expected, subjects)

        self.assertNotIn("session", ents)

        tasks = ents["task"]
        self.assertSetEqual({'mixedgamblestask'}, tasks)

    def test_bidslayout_get(self):
        layout = ancpbids.load_dataset(SYNTHETIC_DIR)
        mask_niftis = layout.query(scope='derivatives',
                                   return_type='filename',
                                   suffix='mask',
                                   extension='.nii',
                                   sub='03',
                                   ses='02',
                                   task='nback',
                                   run=["01", "02"])
        self.assertEqual(4, len(mask_niftis))
        expected_paths = [
            'derivatives/fmriprep/sub-03/ses-02/func/sub-03_ses-02_task-nback_run-01_space-MNI152NLin2009cAsym_desc-brain_mask.nii',
            'derivatives/fmriprep/sub-03/ses-02/func/sub-03_ses-02_task-nback_run-01_space-T1w_desc-brain_mask.nii',
            'derivatives/fmriprep/sub-03/ses-02/func/sub-03_ses-02_task-nback_run-02_space-MNI152NLin2009cAsym_desc-brain_mask.nii',
            'derivatives/fmriprep/sub-03/ses-02/func/sub-03_ses-02_task-nback_run-02_space-T1w_desc-brain_mask.nii',
        ]
        expected_paths = list(map(lambda p: os.path.normpath(os.path.join(SYNTHETIC_DIR, p)), expected_paths))
        for file in expected_paths:
            self.assertTrue(list(filter(lambda p: file == p, mask_niftis)))

    def test_bidslayout_get_entities(self):
        layout = ancpbids.load_dataset(DS005_DIR)
        sorted_entities = layout.query_entities(scope='raw', sort=True)
        # note: 'ds' and 'type' entities are contained in folder 'models' at dataset level, so considered raw data
        self.assertListEqual(['ds', 'run', 'subject', 'task', 'type'], list(sorted_entities.keys()))
        self.assertListEqual([1, 2, 3], sorted_entities['run'])
        self.assertEqual(['%02d' % i for i in range(1, 17)], sorted_entities['subject'])
        self.assertListEqual(['mixedgamblestask'], sorted_entities['task'])

    def test_bidslayout_get_suffixes(self):
        layout = ancpbids.load_dataset(DS005_DIR)
        suffixes = layout.query(target="suffixe")
        self.assertListEqual(['T1w', 'bold', 'dwi', 'events', 'model'], suffixes)

    def test_bidslayout_get_extensions(self):
        layout = ancpbids.load_dataset(DS005_DIR)
        extensions = layout.query(target="extension")
        self.assertListEqual(['.json', '.nii.gz', '.tsv'], extensions)

    def test_bidslayout_get_metadata(self):
        layout = ancpbids.load_dataset(DS005_DIR)
        metadata = layout.get_file("sub-01/func/sub-01_task-mixedgamblestask_run-01_bold.nii.gz").get_metadata(
            include_entities=True)
        self.assertTrue(isinstance(metadata, dict))
        self.assertEqual(2.0, metadata['RepetitionTime'])
        self.assertEqual('mixed-gambles task', metadata['TaskName'])
        self.assertListEqual([0.0, 0.0571, 0.1143, 0.1714, 0.2286, 0.2857], metadata['SliceTiming'])
        self.assertEqual('01', metadata['subject'])
        self.assertEqual('mixedgamblestask', metadata['task'])
        self.assertEqual(1, metadata['run'])

    def test_query_language(self):
        ds = ancpbids.load_dataset(DS005_DIR)
        schema = ds.get_schema()
        file_paths = ds.select(schema.Artifact) \
            .where(all_of(eq(schema.Artifact.suffix, 'bold'),
                          entity(schema, schema.EntityEnum.subject, '02'))) \
            .get_file_paths()
        file_paths = list(file_paths)
        self.assertEqual(3, len(file_paths))


if __name__ == '__main__':
    unittest.main()



import ancpbids
from ..base_test_case import *


class RegressionsTestCase(BaseTestCase):
    def test_get_all_files_from_pipeline(self):
        dataset = ancpbids.load_dataset(DS005_DIR)
        all_derivative_files = dataset.query(scope='derivatives/affine/matrix', return_type='file')
        # that derivatives folder has no valid BIDS files (of type model.Artifact)
        # but ordinary files (if type model.File)
        self.assertEqual(16, len(all_derivative_files))

    def test_report_errors(self):
        dataset = ancpbids.load_dataset(DS005_DIR)
        report = ancpbids.validate_dataset(dataset)
        self.assertTrue(report.has_errors())

    def test_get_entitites_no_scope(self):
        dataset = ancpbids.load_dataset(DS005_DIR)
        entities = dataset.query_entities(scope=None)
        self.assertEqual(['ds', 'type', 'subject', 'task', 'run', 'description'], list(entities.keys()))


if __name__ == '__main__':
    unittest.main()



import importlib
import inspect
import pkgutil
from typing import List

# global plugins registry (list of plugin metadata/settings)
__PLUGINS__ = []


class Plugin:
    """Base class of all plugins.
    """

    def __init__(self, **props):
        self.props = props


class SchemaPlugin(Plugin):
    """A schema plugin may extend/modify a BIDS schema representation module.
    For example, to monkey-patch generated classes.
    """

    def execute(self, schema):
        raise NotImplementedError()


class DatasetPlugin(Plugin):
    """A dataset plugin may enhance an in-memory graph of a dataset.
    """

    def execute(self, dataset):
        raise NotImplementedError()


class FileHandlerPlugin(Plugin):
    """A file handler plugin may register a reader or writer function to allow handling unknown file extensions.
    """

    def execute(self, file_readers_registry, file_writers_registry):
        raise NotImplementedError()


class WritingPlugin(Plugin):
    """A writing plugin may write additional files/folders when a dataset is stored back to file system.
    This may be most interesting to write derivatives to a dataset."""

    def execute(self, dataset, target_dir: str, context_folder=None,
                src_dir: str = None):
        raise NotImplementedError()


class ValidationPlugin(Plugin):
    """A validation plugin may extend the rules to validate a dataset against."""

    class ValidationReport:
        """Contains validation messages (errors/warnings) after a dataset has been validated."""

        def __init__(self):
            self.messages = []

        def error(self, message, offender=None):
            """Adds a new error message to the report.

            Parameters
            ----------
            message:
                the error message to add to the report

            """
            self.messages.append({
                'severity': 'error',
                'offender': offender,
                'message': message
            })

        def warn(self, message, offender=None):
            """Adds a new warning message to the report.

            Parameters
            ----------
            message:
                the warning message to add to the report

            """
            self.messages.append({
                'severity': 'warn',
                'offender': offender,
                'message': message
            })

        def has_errors(self):
            """
            Returns
            -------
            bool
                whether this report contains errors
            """
            return len(self.get_errors()) > 0

        def get_errors(self):
            return list(filter(lambda m: m['severity'] == 'error', self.messages))

    def execute(self, dataset, report: ValidationReport):
        raise NotImplementedError()


def is_valid_plugin(plugin_class):
    """
    Parameters
    ----------
    plugin_class:
        the class to check if known to be a valid plugin class

    Returns
    -------
    bool
        whether the class is considered a valid plugin class
    """
    plugin_types = (SchemaPlugin, DatasetPlugin, WritingPlugin, ValidationPlugin, FileHandlerPlugin)
    return issubclass(plugin_class, plugin_types) and plugin_class not in plugin_types


def load_plugins_by_package(ns_pkg, ranking: int = 1000, **props):
    """Loads all valid plugin classes by the provided package.

    Parameters
    ----------
    ns_pkg:
        the package to scan for plugin classes
    ranking:
        the ranking to use for any detected plugin class
    props
        the properties to assign to the detected plugin classes
    Returns
    -------
    list
        a list of plugin classes or empty if no valid plugin classes found
    """
    mods = [importlib.import_module(name) for finder, name, ispkg in
            pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")]
    for mod in mods:
        for mem in inspect.getmembers(mod, inspect.isclass):
            plugin_class = mem[1]
            if is_valid_plugin(plugin_class):
                register_plugin(plugin_class, ranking, **props)


def register_plugin(plugin_class, ranking: int = 1000, **props):
    """Registers the provided plugin class. If the class is not considered a valid plugin class a ValueError is raised.

    Parameters
    ----------
    plugin_class:
        The plugin class to register.
    ranking:
        The rank to use for the plugin to help prioritize plugins of same type.
        Note that the lower the ranking the higher its prioritization in the processing.
        System level plugins are registered with `ranking = 0`, i.e.
        if you need your plugin to be prioritized over system plugins, use a ranking below 0.
    props
        Additional (static) properties to attach to the provided plugin class.

    """
    if not is_valid_plugin(plugin_class):
        raise ValueError('Invalid plugin class: %s' % plugin_class.__name__)

    __PLUGINS__.append({
        'ranking': ranking,
        'plugin_class': plugin_class,
        'props': props
    })


def get_plugins(plugin_class, **props) -> List[Plugin]:
    """Returns a list of plugin instances matching the provided plugin class and properties.

    Parameters
    ----------
    plugin_class:
        the plugin class to filter by
    props:
        additional filters found in any attached plugin properties

    Returns
    -------
        a list of plugin instances matching the provided plugin class and properties
    """
    plugins = filter(lambda entry: issubclass(entry['plugin_class'], plugin_class), __PLUGINS__)
    plugins = sorted(plugins, key=lambda entry: entry['ranking'])
    # note that a concrete instance of the plugin classes is returned
    return list(map(lambda entry: entry['plugin_class'](**entry['props']), plugins))



from typing import Optional

from ancpbids.plugin import FileHandlerPlugin


def read_yaml(file_path: str, **kwargs):
    import yaml
    with open(file_path, 'r') as stream:
        try:
            return yaml.load(stream, Loader=yaml.FullLoader)
        except:
            return None


def read_json(file_path: str, **kwargs):
    # we cannot use yaml to load json if it contains any TABs for indentation
    import json
    with open(file_path, 'r') as stream:
        try:
            return json.load(stream)
        except:
            return None


def read_plain_text(file_path: str, **kwargs):
    with open(file_path, 'r') as file:
        return file.readlines()


def read_tsv(file_path: str, return_type: Optional[str] = None, **kwargs):
    if return_type == "ndarray":
        import numpy

        return numpy.genfromtxt(
            file_path, delimiter='\t', dtype=None, names=True
        )
    elif return_type == "dataframe":
        import pandas

        return pandas.read_csv(file_path, delimiter='\t')
    else:
        import csv

        with open(file_path) as f:
            return list(csv.DictReader(f, dialect="excel-tab"))


def write_json(file_path: str, contents: dict, **kwargs):
    """Writes the contents as a .json file to the given file path.

    Parameters
    ----------
    file_path:
        The path to the file to store the contents to.
    contents:
        The contents of the target .json file.

    """
    import json
    with open(file_path, 'w') as fp:
        json.dump(contents, fp, indent=2)

def write_txt(file_path: str, contents: dict, **kwargs):
    """Writes the contents as a .txt file to the given file path.

    Parameters
    ----------
    file_path:
        The path to the file to store the contents to.
    contents:
        The contents of the target .txt file.

    """
    with open(file_path, 'w') as fp:
        fp.write(str(contents))


class FilesHandlerPlugin(FileHandlerPlugin):
    def execute(self, file_readers_registry, file_writers_registry):
        file_readers_registry['yaml'] = read_yaml
        file_readers_registry['json'] = read_json
        file_readers_registry['txt'] = read_plain_text
        file_readers_registry['tsv'] = read_tsv

        file_writers_registry['json'] = write_json
        file_writers_registry['txt'] = write_txt



from .model_base import *

VERSION = '1.10.0'
SCHEMA = sys.modules[__name__]


class DatatypeEnum(DatatypeEnum):
    anat = {'value': 'anat', 'display_name': 'Anatomical Magnetic Resonance Imaging',
            'description': 'Magnetic resonance imaging sequences designed to characterize static, anatomical features.\n'}
    beh = {'value': 'beh', 'display_name': 'Behavioral Data', 'description': 'Behavioral data.\n'}
    dwi = {'value': 'dwi', 'display_name': 'Diffusion-Weighted Imaging',
           'description': 'Diffusion-weighted imaging (DWI).\n'}
    eeg = {'value': 'eeg', 'display_name': 'Electroencephalography', 'description': 'Electroencephalography'}
    fmap = {'value': 'fmap', 'display_name': 'Field maps',
            'description': 'MRI scans for estimating B0 inhomogeneity-induced distortions.\n'}
    func = {'value': 'func', 'display_name': 'Task-Based Magnetic Resonance Imaging',
            'description': 'Task (including resting state) imaging data\n'}
    ieeg = {'value': 'ieeg', 'display_name': 'Intracranial electroencephalography',
            'description': 'Intracranial electroencephalography (iEEG) or electrocorticography (ECoG) data\n'}
    meg = {'value': 'meg', 'display_name': 'Magnetoencephalography', 'description': 'Magnetoencephalography'}
    micr = {'value': 'micr', 'display_name': 'Microscopy', 'description': 'Microscopy'}
    motion = {'value': 'motion', 'display_name': 'Motion', 'description': 'Motion data from a tracking system'}
    mrs = {'value': 'mrs', 'display_name': 'Magnetic Resonance Spectroscopy',
           'description': 'Magnetic resonance spectroscopy data'}
    perf = {'value': 'perf', 'display_name': 'Perfusion imaging',
            'description': 'Blood perfusion imaging data, including arterial spin labeling (ASL)\n'}
    pet = {'value': 'pet', 'display_name': 'Positron Emission Tomography',
           'description': 'Positron emission tomography data\n'}
    nirs = {'value': 'nirs', 'display_name': 'Near-Infrared Spectroscopy',
            'description': 'Near-Infrared Spectroscopy data organized around the SNIRF format'}


class ModalityEnum(ModalityEnum):
    mri = {'display_name': 'Magnetic Resonance Imaging', 'description': 'Data acquired with an MRI scanner.\n'}
    eeg = {'display_name': 'Electroencephalography', 'description': 'Data acquired with EEG.\n'}
    ieeg = {'display_name': 'Intracranial Electroencephalography', 'description': 'Data acquired with iEEG.\n'}
    meg = {'display_name': 'Magnetoencephalography', 'description': 'Data acquired with an MEG scanner.\n'}
    beh = {'display_name': 'Behavioral experiments',
           'description': 'Behavioral data acquired without accompanying neuroimaging data.\n'}
    pet = {'display_name': 'Positron Emission Tomography', 'description': 'Data acquired with PET.\n'}
    micr = {'display_name': 'Microscopy', 'description': 'Data acquired with a microscope.\n'}
    motion = {'display_name': 'Motion', 'description': 'Data acquired with Motion-Capture systems.\n'}
    nirs = {'display_name': 'Near-Infrared Spectroscopy', 'description': 'Data acquired with NIRS.'}
    mrs = {'display_name': 'Magnetic Resonance Spectroscopy', 'description': 'Data acquired with MRS.'}


class SuffixEnum(SuffixEnum):
    TwoPE = {'value': '2PE', 'display_name': '2-photon excitation microscopy',
             'description': '2-photon excitation microscopy imaging data\n'}
    BF = {'value': 'BF', 'display_name': 'Bright-field microscopy',
          'description': 'Bright-field microscopy imaging data\n'}
    Chimap = {'value': 'Chimap', 'display_name': 'Quantitative susceptibility map (QSM)',
              'description': 'In parts per million (ppm).\nQSM allows for determining the underlying magnetic susceptibility of tissue\n(Chi)\n([Wang & Liu, 2014](https://doi.org/10.1002/mrm.25358)).\nChi maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
              'unit': 'ppm'}
    CARS = {'value': 'CARS', 'display_name': 'Coherent anti-Stokes Raman spectroscopy',
            'description': 'Coherent anti-Stokes Raman spectroscopy imaging data\n'}
    CONF = {'value': 'CONF', 'display_name': 'Confocal microscopy', 'description': 'Confocal microscopy imaging data\n'}
    DIC = {'value': 'DIC', 'display_name': 'Differential interference contrast microscopy',
           'description': 'Differential interference contrast microscopy imaging data\n'}
    DF = {'value': 'DF', 'display_name': 'Dark-field microscopy', 'description': 'Dark-field microscopy imaging data\n'}
    FLAIR = {'value': 'FLAIR', 'display_name': 'Fluid attenuated inversion recovery image',
             'description': 'In arbitrary units (arbitrary).\nStructural images with predominant T2 contribution (also known as T2-FLAIR),\nin which signal from fluids (for example, CSF) is nulled out by adjusting\ninversion time, coupled with notably long repetition and echo times.\n',
             'unit': 'arbitrary'}
    FLASH = {'value': 'FLASH', 'display_name': 'Fast-Low-Angle-Shot image',
             'description': 'FLASH (Fast-Low-Angle-Shot) is a vendor-specific implementation for spoiled\ngradient echo acquisition.\nIt is commonly used for rapid anatomical imaging and also for many different\nqMRI applications.\nWhen used for a single file, it does not convey any information about the\nimage contrast.\nWhen used in a file collection, it may result in conflicts across filenames of\ndifferent applications.\n**Change:** Removed from suffixes.\n'}
    FLUO = {'value': 'FLUO', 'display_name': 'Fluorescence microscopy',
            'description': 'Fluorescence microscopy imaging data\n'}
    IRT1 = {'value': 'IRT1', 'display_name': 'Inversion recovery T1 mapping',
            'description': 'The IRT1 method involves multiple inversion recovery spin-echo images\nacquired at different inversion times\n([Barral et al. 2010](https://doi.org/10.1002/mrm.22497)).\n'}
    M0map = {'value': 'M0map', 'display_name': 'Equilibrium magnetization (M0) map',
             'description': 'In arbitrary units (arbitrary).\nA common quantitative MRI (qMRI) fitting variable that represents the amount\nof magnetization at thermal equilibrium.\nM0 maps are RECOMMENDED to use this suffix if generated by qMRI applications\n(for example, variable flip angle T1 mapping).\n',
             'unit': 'arbitrary'}
    MEGRE = {'value': 'MEGRE', 'display_name': 'Multi-echo Gradient Recalled Echo',
             'description': 'Anatomical gradient echo images acquired at different echo times.\nPlease note that this suffix is not intended for the logical grouping of\nimages acquired using an Echo Planar Imaging (EPI) readout.\n'}
    MESE = {'value': 'MESE', 'display_name': 'Multi-echo Spin Echo',
            'description': 'The MESE method involves multiple spin echo images acquired at different echo\ntimes and is primarily used for T2 mapping.\nPlease note that this suffix is not intended for the logical grouping of\nimages acquired using an Echo Planar Imaging (EPI) readout.\n'}
    MP2RAGE = {'value': 'MP2RAGE', 'display_name': 'Magnetization Prepared Two Gradient Echoes',
               'description': 'The MP2RAGE method is a special protocol that collects several images at\ndifferent flip angles and inversion times to create a parametric T1map by\ncombining the magnitude and phase images\n([Marques et al. 2010](https://doi.org/10.1016/j.neuroimage.2009.10.002)).\n'}
    MPE = {'value': 'MPE', 'display_name': 'Multi-photon excitation microscopy',
           'description': 'Multi-photon excitation microscopy imaging data\n'}
    MPM = {'value': 'MPM', 'display_name': 'Multi-parametric Mapping',
           'description': 'The MPM approaches (a.k.a hMRI) involves the acquisition of highly-similar\nanatomical images that differ in terms of application of a magnetization\ntransfer RF pulse (MTon or MToff), flip angle and (optionally) echo time and\nmagnitue/phase parts\n([Weiskopf et al. 2013](https://doi.org/10.3389/fnins.2013.00095)).\nSee [here](https://owncloud.gwdg.de/index.php/s/iv2TOQwGy4FGDDZ) for\nsuggested MPM acquisition protocols.\n'}
    MTR = {'value': 'MTR', 'display_name': 'Magnetization Transfer Ratio',
           'description': 'This method is to calculate a semi-quantitative magnetization transfer ratio\nmap.\n'}
    MTRmap = {'value': 'MTRmap', 'display_name': 'Magnetization transfer ratio image',
              'description': 'In arbitrary units (arbitrary).\nMTR maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\nMTRmap intensity values are RECOMMENDED to be represented in percentage in\nthe range of 0-100%.\n',
              'unit': 'arbitrary', 'minValue': 0, 'maxValue': 100}
    MTS = {'value': 'MTS', 'display_name': 'Magnetization transfer saturation',
           'description': 'This method is to calculate a semi-quantitative magnetization transfer\nsaturation index map.\nThe MTS method involves three sets of anatomical images that differ in terms\nof application of a magnetization transfer RF pulse (MTon or MToff) and flip\nangle ([Helms et al. 2008](https://doi.org/10.1002/mrm.21732)).\n'}
    MTVmap = {'value': 'MTVmap', 'display_name': 'Macromolecular tissue volume (MTV) image',
              'description': 'In arbitrary units (arbitrary).\nMTV maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
              'unit': 'arbitrary'}
    MTsat = {'value': 'MTsat', 'display_name': 'Magnetization transfer saturation image',
             'description': 'In arbitrary units (arbitrary).\nMTsat maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
             'unit': 'arbitrary'}
    MWFmap = {'value': 'MWFmap', 'display_name': 'Myelin water fraction image',
              'description': 'In arbitrary units (arbitrary).\nMWF maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\nMWF intensity values are RECOMMENDED to be represented in percentage in the\nrange of 0-100%.\n',
              'unit': 'arbitrary', 'minValue': 0, 'maxValue': 100}
    NLO = {'value': 'NLO', 'display_name': 'Nonlinear optical microscopy',
           'description': 'Nonlinear optical microscopy imaging data\n'}
    OCT = {'value': 'OCT', 'display_name': 'Optical coherence tomography',
           'description': 'Optical coherence tomography imaging data\n'}
    PC = {'value': 'PC', 'display_name': 'Phase-contrast microscopy',
          'description': 'Phase-contrast microscopy imaging data\n'}
    PD = {'value': 'PD', 'display_name': 'Proton density image',
          'description': 'Ambiguous, may refer to a parametric image or to a conventional image.\n**Change:** Replaced by `PDw` or `PDmap`.\n',
          'unit': 'arbitrary'}
    PDT2 = {'value': 'PDT2', 'display_name': 'PD and T2 weighted image',
            'description': 'In arbitrary units (arbitrary).\nA two-volume 4D image, where the volumes are, respectively, PDw and T2w\nimages acquired simultaneously.\nIf separated into 3D volumes, the `PDw` and `T2w` suffixes SHOULD be used instead,\nand an acquisition entity MAY be used to distinguish the images from others with\nthe same suffix, for example, `acq-PDT2_PDw.nii` and `acq-PDT2_T2w.nii`.\n',
            'unit': 'arbitrary'}
    PDmap = {'value': 'PDmap', 'display_name': 'Proton density image',
             'description': 'In arbitrary units (arbitrary).\nPD maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
             'unit': 'arbitrary'}
    PDw = {'value': 'PDw', 'display_name': 'Proton density (PD) weighted image',
           'description': 'In arbitrary units (arbitrary).\nThe contrast of these images is mainly determined by spatial variations in\nthe spin density (1H) of the imaged specimen.\nThis contrast is achieved at short echo times and long repetition times;\nfor gradient echo, this weighting is also possible with a short TR (TR<<T1) and a small flip angle.\n',
           'unit': 'arbitrary'}
    PLI = {'value': 'PLI', 'display_name': 'Polarized-light microscopy',
           'description': 'Polarized-light microscopy imaging data\n'}
    R1map = {'value': 'R1map', 'display_name': 'Longitudinal relaxation rate image',
             'description': 'In seconds<sup>-1</sup> (1/s).\nR1 maps (R1 = 1/T1) are REQUIRED to use this suffix regardless of the method\nused to generate them.\n',
             'unit': '1/s'}
    R2map = {'value': 'R2map', 'display_name': 'True transverse relaxation rate image',
             'description': 'In seconds<sup>-1</sup> (1/s).\nR2 maps (R2 = 1/T2) are REQUIRED to use this suffix regardless of the method\nused to generate them.\n',
             'unit': '1/s'}
    R2starmap = {'value': 'R2starmap', 'display_name': 'Observed transverse relaxation rate image',
                 'description': 'In seconds<sup>-1</sup> (1/s).\nR2-star maps (R2star = 1/T2star) are REQUIRED to use this suffix regardless\nof the method used to generate them.\n',
                 'unit': '1/s'}
    RB1COR = {'value': 'RB1COR', 'display_name': 'RB1COR',
              'description': 'Low resolution images acquired by the body coil\n(in the gantry of the scanner) and the head coil using identical acquisition\nparameters to generate a combined sensitivity map as described in\n[Papp et al. (2016)](https://doi.org/10.1002/mrm.26058).\n'}
    RB1map = {'value': 'RB1map', 'display_name': 'RF receive sensitivity map',
              'description': 'In arbitrary units (arbitrary).\nRadio frequency (RF) receive (B1-) sensitivity maps are REQUIRED to use this\nsuffix regardless of the method used to generate them.\nRB1map intensity values are RECOMMENDED to be represented as percent\nmultiplicative factors such that Amplitude<sub>effective</sub> =\nB1-<sub>intensity</sub>\\*Amplitude<sub>ideal</sub>.\n',
              'unit': 'arbitrary'}
    S0map = {'value': 'S0map', 'display_name': 'Observed signal amplitude (S0) image',
             'description': 'In arbitrary units (arbitrary).\nFor a multi-echo (typically fMRI) sequence, S0 maps index the baseline signal\nbefore exponential (T2-star) signal decay.\nIn other words: the exponential of the intercept for a linear decay model\nacross log-transformed echos. For more information, please see, for example,\n[the tedana documentation](https://tedana.readthedocs.io/en/latest/\\\napproach.html#monoexponential-decay-model-fit).\nS0 maps are RECOMMENDED to use this suffix if derived from an ME-FMRI dataset.\n'}
    SEM = {'value': 'SEM', 'display_name': 'Scanning electron microscopy',
           'description': 'Scanning electron microscopy imaging data\n'}
    SPIM = {'value': 'SPIM', 'display_name': 'Selective plane illumination microscopy',
            'description': 'Selective plane illumination microscopy imaging data\n'}
    SR = {'value': 'SR', 'display_name': 'Super-resolution microscopy',
          'description': 'Super-resolution microscopy imaging data\n'}
    T1map = {'value': 'T1map', 'display_name': 'Longitudinal relaxation time image',
             'description': 'In seconds (s).\nT1 maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\nSee [this interactive book on T1 mapping](https://qmrlab.org/t1_book/intro)\nfor further reading on T1-mapping.\n',
             'unit': 's'}
    T1rho = {'value': 'T1rho', 'display_name': 'T1 in rotating frame (T1 rho) image',
             'description': 'In seconds (s).\nT1-rho maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
             'unit': 's'}
    T1w = {'value': 'T1w', 'display_name': 'T1-weighted image',
           'description': 'In arbitrary units (arbitrary).\nThe contrast of these images is mainly determined by spatial variations in\nthe longitudinal relaxation time of the imaged specimen.\nIn spin-echo sequences this contrast is achieved at relatively short\nrepetition and echo times.\nTo achieve this weighting in gradient-echo images, again, short repetition\nand echo times are selected; however, at relatively large flip angles.\nAnother common approach to increase T1 weighting in gradient-echo images is\nto add an inversion preparation block to the beginning of the imaging\nsequence (for example, `TurboFLASH` or `MP-RAGE`).\n',
           'unit': 'arbitrary'}
    T2map = {'value': 'T2map', 'display_name': 'True transverse relaxation time image',
             'description': 'In seconds (s).\nT2 maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
             'unit': 's'}
    T2star = {'value': 'T2star', 'display_name': 'T2\\* image',
              'description': 'Ambiguous, may refer to a parametric image or to a conventional image.\n**Change:** Replaced by `T2starw` or `T2starmap`.\n',
              'anyOf': [{'unit': 'arbitrary'}, {'unit': 's'}]}
    T2starmap = {'value': 'T2starmap', 'display_name': 'Observed transverse relaxation time image',
                 'description': 'In seconds (s).\nT2-star maps are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\n',
                 'unit': 's'}
    T2starw = {'value': 'T2starw', 'display_name': 'T2star weighted image',
               'description': 'In arbitrary units (arbitrary).\nThe contrast of these images is mainly determined by spatial variations in\nthe (observed) transverse relaxation time of the imaged specimen.\nIn spin-echo sequences, this effect is negated as the excitation is followed\nby an inversion pulse.\nThe contrast of gradient-echo images natively depends on T2-star effects.\nHowever, for T2-star variation to dominate the image contrast,\ngradient-echo acquisitions are carried out at long repetition and echo times,\nand at small flip angles.\n',
               'unit': 'arbitrary'}
    T2w = {'value': 'T2w', 'display_name': 'T2-weighted image',
           'description': 'In arbitrary units (arbitrary).\nThe contrast of these images is mainly determined by spatial variations in\nthe (true) transverse relaxation time of the imaged specimen.\nIn spin-echo sequences this contrast is achieved at relatively long\nrepetition and echo times.\nGenerally, gradient echo sequences are not the most suitable option for\nachieving T2 weighting, as their contrast natively depends on T2-star rather\nthan on T2.\n',
           'unit': 'arbitrary'}
    TB1AFI = {'value': 'TB1AFI', 'display_name': 'TB1AFI',
              'description': 'This method ([Yarnykh 2007](https://doi.org/10.1002/mrm.21120))\ncalculates a B1<sup>+</sup> map from two images acquired at interleaved (two)\nTRs with identical RF pulses using a steady-state sequence.\n'}
    TB1DAM = {'value': 'TB1DAM', 'display_name': 'TB1DAM',
              'description': 'The double-angle B1<sup>+</sup> method\n([Insko and Bolinger 1993](https://doi.org/10.1006/jmra.1993.1133)) is based\non the calculation of the actual angles from signal ratios,\ncollected by two acquisitions at different nominal excitation flip angles.\nCommon sequence types for this application include spin echo and echo planar\nimaging.\n'}
    TB1EPI = {'value': 'TB1EPI', 'display_name': 'TB1EPI',
              'description': 'This B1<sup>+</sup> mapping method\n([Jiru and Klose 2006](https://doi.org/10.1002/mrm.21083)) is based on two\nEPI readouts to acquire spin echo (SE) and stimulated echo (STE) images at\nmultiple flip angles in one sequence, used in the calculation of deviations\nfrom the nominal flip angle.\n'}
    TB1RFM = {'value': 'TB1RFM', 'display_name': 'TB1RFM',
              'description': 'The result of a Siemens `rf_map` product sequence.\nThis sequence produces two images.\nThe first image appears like an anatomical image and the second output is a\nscaled flip angle map.\n'}
    TB1SRGE = {'value': 'TB1SRGE', 'display_name': 'TB1SRGE',
               'description': 'Saturation-prepared with 2 rapid gradient echoes (SA2RAGE) uses a ratio of\ntwo saturation recovery images with different time delays,\nand a simulated look-up table to estimate B1+\n([Eggenschwiler et al. 2011](https://doi.org/10.1002/mrm.23145)).\nThis sequence can also be used in conjunction with MP2RAGE T1 mapping to\niteratively improve B1+ and T1 map estimation\n([Marques & Gruetter 2013](https://doi.org/10.1371/journal.pone.0069294)).\n'}
    TB1TFL = {'value': 'TB1TFL', 'display_name': 'TB1TFL',
              'description': 'The result of a Siemens `tfl_b1_map` product sequence.\nThis sequence produces two images.\nThe first image appears like an anatomical image and the second output is a\nscaled flip angle map.\n'}
    TB1map = {'value': 'TB1map', 'display_name': 'RF transmit field image',
              'description': 'In arbitrary units (arbitrary).\nRadio frequency (RF) transmit (B1+) field maps are REQUIRED to use this\nsuffix regardless of the method used to generate them.\nTB1map intensity values are RECOMMENDED to be represented as percent\nmultiplicative factors such that FlipAngle<sub>effective</sub> =\nB1+<sub>intensity</sub>\\*FlipAngle<sub>nominal</sub> .\n',
              'unit': 'arbitrary'}
    TEM = {'value': 'TEM', 'display_name': 'Transmission electron microscopy',
           'description': 'Transmission electron microscopy imaging data\n'}
    UNIT1 = {'value': 'UNIT1', 'display_name': 'Homogeneous (flat) T1-weighted MP2RAGE image',
             'description': 'In arbitrary units (arbitrary).\nUNIT1 images are REQUIRED to use this suffix regardless of the method used to\ngenerate them.\nNote that although this image is T1-weighted, regions without MR signal will\ncontain white salt-and-pepper noise that most segmentation algorithms will\nfail on.\nTherefore, it is important to dissociate it from `T1w`.\nPlease see [`MP2RAGE` specific notes](SPEC_ROOT/appendices/qmri.md#unit1-images)\nin the qMRI appendix for further information.\n'}
    VFA = {'value': 'VFA', 'display_name': 'Variable flip angle',
           'description': 'The VFA method involves at least two spoiled gradient echo (SPGR) of\nsteady-state free precession (SSFP) images acquired at different flip angles.\nDepending on the provided metadata fields and the sequence type,\ndata may be eligible for DESPOT1, DESPOT2 and their variants\n([Deoni et al. 2005](https://doi.org/10.1002/mrm.20314)).\n'}
    angio = {'value': 'angio', 'display_name': 'Angiogram',
             'description': 'Magnetic resonance angiography sequences focus on enhancing the contrast of\nblood vessels (generally arteries, but sometimes veins) against other tissue\ntypes.\n'}
    asl = {'value': 'asl', 'display_name': 'Arterial Spin Labeling',
           'description': 'The complete ASL time series stored as a 4D NIfTI file in the original\nacquisition order, with possible volume types including: control, label,\nm0scan, deltam, cbf.\n'}
    aslcontext = {'value': 'aslcontext', 'display_name': 'Arterial Spin Labeling Context',
                  'description': 'A TSV file defining the image types for volumes in an associated ASL file.\n'}
    asllabeling = {'value': 'asllabeling', 'display_name': 'ASL Labeling Screenshot',
                   'description': 'A deidentified screenshot of the planning of the labeling slab/plane\nwith respect to the imaging slab or slices.\nThis screenshot is based on DICOM macro C.8.13.5.14.\n'}
    beh = {'value': 'beh', 'display_name': 'Behavioral recording',
           'description': 'Behavioral recordings from tasks.\nThese files are similar to events files, but do not include the `"onset"` and\n`"duration"` columns that are mandatory for events files.\n'}
    blood = {'value': 'blood', 'display_name': 'Blood recording data',
             'description': 'Blood measurements of radioactivity stored in\n[tabular files](SPEC_ROOT/common-principles.md#tabular-files)\nand located in the `pet/` directory along with the corresponding PET data.\n'}
    bold = {'value': 'bold', 'display_name': 'Blood-Oxygen-Level Dependent image',
            'description': 'Blood-Oxygen-Level Dependent contrast (specialized T2\\* weighting)\n'}
    cbv = {'value': 'cbv', 'display_name': 'Cerebral blood volume image',
           'description': 'Cerebral Blood Volume contrast (specialized T2\\* weighting or difference between T1 weighted images)\n'}
    channels = {'value': 'channels', 'display_name': 'Channels File', 'description': 'Channel information.\n'}
    coordsystem = {'value': 'coordsystem', 'display_name': 'Coordinate System File',
                   'description': 'A JSON document specifying the coordinate system(s) used for the MEG, EEG,\nhead localization coils, and anatomical landmarks.\n'}
    defacemask = {'value': 'defacemask', 'display_name': 'Defacing Mask',
                  'description': 'A binary mask that was used to remove facial features from an anatomical MRI\nimage.\n'}
    descriptions = {'value': 'descriptions', 'display_name': 'Description Entity Definitions',
                    'description': 'A TSV file describing labels found for the `desc` entity in a Derivatives dataset.\n\nThis file MAY be located at the root of the derivative dataset,\nor at the subject or session level.\n\nThe use of `descriptions.tsv` files together with the desc entity are helpful to\ndocument how files are generated,\neven if their use may not be sufficient to provide full computational reproducibility.\n'}
    dseg = {'value': 'dseg', 'display_name': 'Discrete Segmentation',
            'description': 'A discrete segmentation.\n\nThis suffix may only be used in derivative datasets.\n'}
    dwi = {'value': 'dwi', 'display_name': 'Diffusion-weighted image',
           'description': 'Diffusion-weighted imaging contrast (specialized T2 weighting).\n'}
    eeg = {'value': 'eeg', 'display_name': 'Electroencephalography',
           'description': 'Electroencephalography recording data.\n'}
    electrodes = {'value': 'electrodes', 'display_name': 'Electrodes',
                  'description': 'File that gives the location of (i)EEG electrodes.\n'}
    epi = {'value': 'epi', 'display_name': 'EPI',
           'description': 'The phase-encoding polarity (PEpolar) technique combines two or more Spin Echo\nEPI scans with different phase encoding directions to estimate the underlying\ninhomogeneity/deformation map.\n'}
    events = {'value': 'events', 'display_name': 'Events',
              'description': 'Event timing information from a behavioral task.\n'}
    fieldmap = {'value': 'fieldmap', 'display_name': 'Fieldmap',
                'description': 'Some MR schemes such as spiral-echo imaging (SEI) sequences are able to\ndirectly provide maps of the *B<sub>0</sub>* field inhomogeneity.\n'}
    headshape = {'value': 'headshape', 'display_name': 'Headshape File',
                 'description': 'The 3-D locations of points that describe the head shape and/or electrode\nlocations can be digitized and stored in separate files.\n'}
    XPCT = {'value': 'XPCT', 'display_name': 'X-ray Phase-Contrast Tomography',
            'description': 'X-ray phase-contrast tomography imaging data\n'}
    ieeg = {'value': 'ieeg', 'display_name': 'Intracranial Electroencephalography',
            'description': 'Intracranial electroencephalography recording data.\n'}
    inplaneT1 = {'value': 'inplaneT1', 'display_name': 'Inplane T1',
                 'description': 'In arbitrary units (arbitrary).\nT1 weighted structural image matched to a functional (task) image.\n',
                 'unit': 'arbitrary'}
    inplaneT2 = {'value': 'inplaneT2', 'display_name': 'Inplane T2',
                 'description': 'In arbitrary units (arbitrary).\nT2 weighted structural image matched to a functional (task) image.\n',
                 'unit': 'arbitrary'}
    m0scan = {'value': 'm0scan', 'display_name': 'M0 image',
              'description': 'The M0 image is a calibration image, used to estimate the equilibrium\nmagnetization of blood.\n'}
    magnitude = {'value': 'magnitude', 'display_name': 'Magnitude',
                 'description': 'Field-mapping MR schemes such as gradient-recalled echo (GRE) generate a\nMagnitude image to be used for anatomical reference.\nRequires the existence of Phase, Phase-difference or Fieldmap maps.\n'}
    magnitude1 = {'value': 'magnitude1', 'display_name': 'Magnitude',
                  'description': 'Magnitude map generated by GRE or similar schemes, associated with the first\necho in the sequence.\n'}
    magnitude2 = {'value': 'magnitude2', 'display_name': 'Magnitude',
                  'description': 'Magnitude map generated by GRE or similar schemes, associated with the second\necho in the sequence.\n'}
    markers = {'value': 'markers', 'display_name': 'MEG Sensor Coil Positions',
               'description': 'Another manufacturer-specific detail pertains to the KIT/Yokogawa/Ricoh\nsystem, which saves the MEG sensor coil positions in a separate file with two\npossible filename extensions  (`.sqd`, `.mrk`).\nFor these files, the `markers` suffix MUST be used.\nFor example: `sub-01_task-nback_markers.sqd`\n'}
    mask = {'value': 'mask', 'display_name': 'Binary Mask',
            'description': 'A binary mask that functions as a discrete "label" for a single structure.\n\nThis suffix may only be used in derivative datasets.\n'}
    meg = {'value': 'meg', 'display_name': 'Magnetoencephalography',
           'description': 'Unprocessed MEG data stored in the native file format of the MEG instrument\nwith which the data was collected.\n'}
    motion = {'value': 'motion', 'display_name': 'Motion',
              'description': 'Data recorded from a tracking system store.\n'}
    mrsi = {'value': 'mrsi', 'display_name': 'Magnetic resonance spectroscopy imaging',
            'description': 'MRS acquisitions where additional imaging gradients are used\nto detect the MR signal from 1, 2, or 3 spatial dimensions.\n'}
    mrsref = {'value': 'mrsref', 'display_name': 'MRS reference acquisition',
              'description': 'An MRS acquisition collected to serve as a concentration reference for absolute quantification\nor as a calibration reference for preprocessing (for example, eddy-current correction).\n'}
    nirs = {'value': 'nirs', 'display_name': 'Near Infrared Spectroscopy',
            'description': 'Data associated with a Shared Near Infrared Spectroscopy Format file.'}
    noRF = {'value': 'noRF', 'display_name': 'No Radio Frequency Excitation Scan',
            'description': 'An MR noise scan often acquired alongside another MR scan.\n\nThese noise scans are acquired by disabling the radio frequency excitation,\nwhile maintaining all other parameters from the associated scan.\nThese can be acquired using some other sequence with RF excitation removed\nor as part of sequences that turns off RF at the beginning or end of a run.\nFor example a `bold` scan where a few volumes are acquired with RF off.\nThere is no consistent DICOM denotation for these scans.\n\nIf multiple sequences with different suffixes but otherwise the same entities have `noRF`\nsequences, they SHOULD be disambiguated using the modality entity\n(for example, `mod-bold_noRF.nii.gz` and `mod-cbv_noRF.nii`).\n\n`noRF` are typically used to estimate properties of the scanner and\nsequence that are independent from the object being imaged.\n'}
    optodes = {'value': 'optodes', 'display_name': 'Optodes',
               'description': 'Either a light emitting device, sometimes called a transmitter, or a photoelectric transducer, sometimes called a\nreceiver.\n'}
    pet = {'value': 'pet', 'display_name': 'Positron Emission Tomography',
           'description': 'PET imaging data SHOULD be stored in 4D\n(or 3D, if only one volume was acquired) NIfTI files with the `_pet` suffix.\nVolumes MUST be stored in chronological order\n(the order they were acquired in).\n'}
    phase = {'value': 'phase', 'display_name': 'Phase image',
             'description': '[DEPRECATED](SPEC_ROOT/common-principles.md#definitions).\nPhase information associated with magnitude information stored in BOLD\ncontrast.\nThis suffix should be replaced by the\n[`part-phase`](SPEC_ROOT/appendices/entities.md#part)\nin conjunction with the `bold` suffix.\n',
             'anyOf': [{'unit': 'arbitrary'}, {'unit': 'rad'}]}
    phase1 = {'value': 'phase1', 'display_name': 'Phase',
              'description': 'Phase map generated by GRE or similar schemes, associated with the first\necho in the sequence.\n'}
    phase2 = {'value': 'phase2', 'display_name': 'Phase',
              'description': 'Phase map generated by GRE or similar schemes, associated with the second\necho in the sequence.\n'}
    phasediff = {'value': 'phasediff', 'display_name': 'Phase-difference',
                 'description': 'Some scanners subtract the `phase1` from the `phase2` map and generate a\nunique `phasediff` file.\nFor instance, this is a common output for the built-in fieldmap sequence of\nSiemens scanners.\n'}
    photo = {'value': 'photo', 'display_name': 'Photo File',
             'description': 'Photos of the anatomical landmarks, head localization coils or tissue sample.\n'}
    physio = {'value': 'physio', 'display_name': 'Physiological recording',
              'description': 'Physiological recordings such as cardiac and respiratory signals.\n'}
    probseg = {'value': 'probseg', 'display_name': 'Probabilistic Segmentation',
               'description': 'A probabilistic segmentation.\n\nThis suffix may only be used in derivative datasets.\n'}
    sbref = {'value': 'sbref', 'display_name': 'Single-band reference image',
             'description': 'Single-band reference for one or more multi-band `dwi` images.\n'}
    scans = {'value': 'scans', 'display_name': 'Scans file',
             'description': 'The purpose of this file is to describe timing and other properties of each imaging acquisition\nsequence (each run file) within one session.\nEach neural recording file SHOULD be described by exactly one row. Some recordings consist of\nmultiple parts, that span several files, for example through echo-, part-, or split- entities.\nSuch recordings MUST be documented with one row per file.\nRelative paths to files should be used under a compulsory filename header.\nIf acquisition time is included it should be listed under the acq_time header.\nAcquisition time refers to when the first data point in each run was acquired.\nFurthermore, if this header is provided, the acquisition times of all files that belong to a\nrecording MUST be identical.\nDatetime should be expressed as described in Units.\nAdditional fields can include external behavioral measures relevant to the scan.\nFor example vigilance questionnaire score administered after a resting state scan.\nAll such included additional fields SHOULD be documented in an accompanying _scans.json file\nthat describes these fields in detail (see Tabular files).\n'}
    sessions = {'value': 'sessions', 'display_name': 'Sessions file',
                'description': 'In case of multiple sessions there is an option of adding additional sessions.tsv files\ndescribing variables changing between sessions.\nIn such case one file per participant SHOULD be added.\nThese files MUST include a session_id column and describe each session by one and only one row.\nColumn names in sessions.tsv files MUST be different from group level participant key column\nnames in the participants.tsv file.\n'}
    stim = {'value': 'stim', 'display_name': 'Continuous recording',
            'description': 'Continuous measures, such as parameters of a film or audio stimulus.\n'}
    svs = {'value': 'svs', 'display_name': 'Single-voxel spectroscopy',
           'description': 'MRS acquisitions where the detected MR signal is spatially localized to a single volume.\n'}
    uCT = {'value': 'uCT', 'display_name': 'Micro-CT', 'description': 'Micro-CT imaging data\n'}
    unloc = {'value': 'unloc', 'display_name': 'Unlocalized spectroscopy',
             'description': 'MRS acquisitions run without localization.\nThis includes signals detected using coil sensitivity only.\n'}


class EntityEnum(EntityEnum):
    subject = {'name': 'sub', 'display_name': 'Subject',
               'description': 'A person or animal participating in the study.\n', 'type': 'string', 'format': 'label'}
    session = {'name': 'ses', 'display_name': 'Session',
               'description': "A logical grouping of neuroimaging and behavioral data consistent across subjects.\nSession can (but doesn't have to) be synonymous to a visit in a longitudinal study.\nIn general, subjects will stay in the scanner during one session.\nHowever, for example, if a subject has to leave the scanner room and then\nbe re-positioned on the scanner bed, the set of MRI acquisitions will still\nbe considered as a session and match sessions acquired in other subjects.\nSimilarly, in situations where different data types are obtained over\nseveral visits (for example fMRI on one day followed by DWI the day after)\nthose can be grouped in one session.\n\nDefining multiple sessions is appropriate when several identical or similar\ndata acquisitions are planned and performed on all -or most- subjects,\noften in the case of some intervention between sessions\n(for example, training).\n",
               'type': 'string', 'format': 'label'}
    sample = {'name': 'sample', 'display_name': 'Sample',
              'description': 'A sample pertaining to a subject such as tissue, primary cell or cell-free sample.\nThe `sample-<label>` entity is used to distinguish between different samples from the same subject.\nThe label MUST be unique per subject and is RECOMMENDED to be unique throughout the dataset.\n',
              'type': 'string', 'format': 'label'}
    task = {'name': 'task', 'display_name': 'Task',
            'description': 'A set of structured activities performed by the participant.\nTasks are usually accompanied by stimuli and responses, and can greatly vary in complexity.\n\nIn the context of brain scanning, a task is always tied to one data acquisition.\nTherefore, even if during one acquisition the subject performed multiple conceptually different behaviors\n(with different sets of instructions) they will be considered one (combined) task.\n\nWhile tasks may be repeated across multiple acquisitions,\na given task may have different sets of stimuli (for example, randomized order) and participant responses\nacross subjects, sessions, and runs.\n\nThe `task-<label>` MUST be consistent across subjects and sessions.\n\nFiles with the `task-<label>` entity SHOULD have an associated\n[events file](SPEC_ROOT/modality-specific-files/task-events.md#task-events),\nas well as certain metadata fields in the associated JSON file.\n\nFor the purpose of this specification we consider the so-called "resting state" a task,\nalthough events files are not expected for resting state data.\nAdditionally, a common convention in the specification is to include the word "rest" in\nthe `task` label for resting state files (for example, `task-rest`).\n',
            'type': 'string', 'format': 'label'}
    tracksys = {'name': 'tracksys', 'display_name': 'Tracking System',
                'description': 'The `tracksys-<label>` entity can be used as a key-value pair\nto label *_motion.tsv and *_motion.json files.\nIt can also be used to label *_channel.tsv or *_events.tsv files\nwhen they belong to a specific tracking system.\n\nThis entity corresponds to the `"TrackingSystemName"` metadata field in a *_motion.json file.\n`tracksys-<label>` entity is a concise string whereas `"TrackingSystemName"`\nmay be longer and more human readable.\n',
                'type': 'string', 'format': 'label'}
    acquisition = {'name': 'acq', 'display_name': 'Acquisition',
                   'description': 'The `acq-<label>` entity corresponds to a custom label the user MAY use to distinguish\na different set of parameters used for acquiring the same modality.\n\nFor example, this should be used when a study includes two T1w images -\none full brain low resolution and one restricted field of view but high resolution.\nIn such case two files could have the following names:\n`sub-01_acq-highres_T1w.nii.gz` and `sub-01_acq-lowres_T1w.nii.gz`;\nhowever, the user is free to choose any other label than `highres` and `lowres` as long\nas they are consistent across subjects and sessions.\n\nIn case different sequences are used to record the same modality\n(for example, `RARE` and `FLASH` for T1w)\nthis field can also be used to make that distinction.\nThe level of detail at which the distinction is made\n(for example, just between `RARE` and `FLASH`, or between `RARE`, `FLASH`, and `FLASHsubsampled`)\nremains at the discretion of the researcher.\n',
                   'type': 'string', 'format': 'label'}
    nucleus = {'name': 'nuc', 'display_name': 'Nucleus',
               'description': 'The `nuc-<label>` entity can be used to distinguish acquisitions tuned\nto detect different nuclei.\nThe label is the name of the nucleus or nuclei, which corresponds to DICOM Tag `0018, 9100`.\nIf present in the filename, `"ResonantNucleus"` MUST also be included in\nthe associated metadata.\n',
               'type': 'string', 'format': 'label'}
    volume = {'name': 'voi', 'display_name': 'Volume of Interest',
              'description': 'The `voi-<label>` entity can be used to distinguish acquisitions localized to different regions.\nThe label SHOULD be the name of the body region or part scanned.\nIf used, the fields `"BodyPart"` and `"BodyPartDetails"` MUST be defined in the JSON file.\n`BodyPartDetailsOntology` is OPTIONAL to also include.\n',
              'type': 'string', 'format': 'label'}
    ceagent = {'name': 'ce', 'display_name': 'Contrast Enhancing Agent',
               'description': 'The `ce-<label>` entity can be used to distinguish sequences using different contrast enhanced images.\nThe label is the name of the contrast agent.\n\nThis entity represents the `"ContrastBolusIngredient"` metadata field.\nTherefore, if the `ce-<label>` entity is present in a filename,\n`"ContrastBolusIngredient"` MAY also be added in the JSON file, with the same label.\n',
               'type': 'string', 'format': 'label'}
    tracer = {'name': 'trc', 'display_name': 'Tracer',
              'description': 'The `trc-<label>` entity can be used to distinguish sequences using different tracers.\n\nThis entity represents the `"TracerName"` metadata field.\nTherefore, if the `trc-<label>` entity is present in a filename,\n`"TracerName"` MUST be defined in the associated metadata.\nPlease note that the `<label>` does not need to match the actual value of the field.\n',
              'type': 'string', 'format': 'label'}
    stain = {'name': 'stain', 'display_name': 'Stain',
             'description': 'The `stain-<label>` key/pair values can be used to distinguish image files\nfrom the same sample using different stains or antibodies for contrast enhancement.\n\nThis entity represents the `"SampleStaining"` metadata field.\nTherefore, if the `stain-<label>` entity is present in a filename,\n`"SampleStaining"` SHOULD be defined in the associated metadata,\nalthough the label may be different.\n\nDescriptions of antibodies SHOULD also be indicated in the `"SamplePrimaryAntibodies"`\nand/or `"SampleSecondaryAntibodies"` metadata fields, as appropriate.\n',
             'type': 'string', 'format': 'label'}
    reconstruction = {'name': 'rec', 'display_name': 'Reconstruction',
                      'description': 'The `rec-<label>` entity can be used to distinguish different reconstruction algorithms\n(for example, `MoCo` for the ones using motion correction).\n',
                      'type': 'string', 'format': 'label'}
    direction = {'name': 'dir', 'display_name': 'Phase-Encoding Direction',
                 'description': 'The `dir-<label>` entity can be set to an arbitrary alphanumeric label\n(for example, `dir-LR` or `dir-AP`)\nto distinguish different phase-encoding directions.\n\nThis entity represents the `"PhaseEncodingDirection"` metadata field.\nTherefore, if the `dir-<label>` entity is present in a filename,\n`"PhaseEncodingDirection"` MUST be defined in the associated metadata.\nPlease note that the `<label>` does not need to match the actual value of the field.\n',
                 'type': 'string', 'format': 'label'}
    run = {'name': 'run', 'display_name': 'Run',
           'description': 'The `run-<index>` entity is used to distinguish separate data acquisitions with the same acquisition parameters\nand (other) entities.\n\nIf several data acquisitions (for example, MRI scans or EEG recordings)\nwith the same acquisition parameters are acquired in the same session,\nthey MUST be indexed with the [`run-<index>`](SPEC_ROOT/appendices/entities.md#run) entity:\n`_run-1`, `_run-2`, `_run-3`, and so on\n(only nonnegative integers are allowed as run indices).\n\nIf different entities apply,\nsuch as a different session indicated by [`ses-<label>`][SPEC_ROOT/appendices/entities.md#ses),\nor different acquisition parameters indicated by\n[`acq-<label>`](SPEC_ROOT/appendices/entities.md#acq),\nthen `run` is not needed to distinguish the scans and MAY be omitted.\n',
           'type': 'string', 'format': 'index'}
    modality = {'name': 'mod', 'display_name': 'Corresponding Modality',
                'description': 'The `mod-<label>` entity corresponds to modality label for defacing\nmasks, for example, T1w, inplaneT1, referenced by a defacemask image.\nFor example, `sub-01_mod-T1w_defacemask.nii.gz`.\n',
                'type': 'string', 'format': 'label'}
    echo = {'name': 'echo', 'display_name': 'Echo',
            'description': 'If files belonging to an entity-linked file collection are acquired at different\necho times, the `echo-<index>` entity MUST be used to distinguish individual files.\n\nThis entity represents the `"EchoTime"` metadata field.\nTherefore, if the `echo-<index>` entity is present in a filename,\n`"EchoTime"` MUST be defined in the associated metadata.\nPlease note that the `<index>` denotes the number/index (in the form of a nonnegative integer),\nnot the `"EchoTime"` value of the separate JSON file.\n',
            'type': 'string', 'format': 'index'}
    flip = {'name': 'flip', 'display_name': 'Flip Angle',
            'description': 'If files belonging to an entity-linked file collection are acquired at different\nflip angles, the `_flip-<index>` entity pair MUST be used to distinguish\nindividual files.\n\nThis entity represents the `"FlipAngle"` metadata field.\nTherefore, if the `flip-<index>` entity is present in a filename,\n`"FlipAngle"` MUST be defined in the associated metadata.\nPlease note that the `<index>` denotes the number/index (in the form of a nonnegative integer),\nnot the `"FlipAngle"` value of the separate JSON file.\n',
            'type': 'string', 'format': 'index'}
    inversion = {'name': 'inv', 'display_name': 'Inversion Time',
                 'description': 'If files belonging to an entity-linked file collection are acquired at different inversion times,\nthe `inv-<index>` entity MUST be used to distinguish individual files.\n\nThis entity represents the `"InversionTime` metadata field.\nTherefore, if the `inv-<index>` entity is present in a filename,\n`"InversionTime"` MUST be defined in the associated metadata.\nPlease note that the `<index>` denotes the number/index (in the form of a nonnegative integer),\nnot the `"InversionTime"` value of the separate JSON file.\n',
                 'type': 'string', 'format': 'index'}
    mtransfer = {'name': 'mt', 'display_name': 'Magnetization Transfer',
                 'description': 'If files belonging to an entity-linked file collection are acquired at different\nmagnetization transfer (MT) states, the `_mt-<label>` entity MUST be used to\ndistinguish individual files.\n\nThis entity represents the `"MTState"` metadata field.\nTherefore, if the `mt-<label>` entity is present in a filename,\n`"MTState"` MUST be defined in the associated metadata.\nAllowed label values for this entity are `on` and `off`,\nfor images acquired in presence and absence of an MT pulse, respectively.\n',
                 'type': 'string', 'format': 'label', 'enum': ['on', 'off']}
    part = {'name': 'part', 'display_name': 'Part',
            'description': 'This entity is used to indicate which component of the complex\nrepresentation of the MRI signal is represented in voxel data.\nThe `part-<label>` entity is associated with the DICOM Tag\n`0008, 9208`.\nAllowed label values for this entity are `phase`, `mag`, `real` and `imag`,\nwhich are typically used in `part-mag`/`part-phase` or\n`part-real`/`part-imag` pairs of files.\n\nPhase images MAY be in radians or in arbitrary units.\nThe sidecar JSON file MUST include the `"Units"` of the `phase` image.\nThe possible options are `"rad"` or `"arbitrary"`.\n\nWhen there is only a magnitude image of a given type, the `part` entity MAY be\nomitted.\n',
            'type': 'string', 'format': 'label', 'enum': ['mag', 'phase', 'real', 'imag']}
    processing = {'name': 'proc', 'display_name': 'Processed (on device)',
                  'description': "The proc label is analogous to rec for MR and denotes a variant of\na file that was a result of particular processing performed on the device.\n\nThis is useful for files produced in particular by Neuromag/Elekta/MEGIN's\nMaxFilter (for example, `sss`, `tsss`, `trans`, `quat` or `mc`),\nwhich some installations impose to be run on raw data because of active\nshielding software corrections before the MEG data can actually be\nexploited.\n",
                  'type': 'string', 'format': 'label'}
    hemisphere = {'name': 'hemi', 'display_name': 'Hemisphere',
                  'description': 'The `hemi-<label>` entity indicates which hemibrain is described by the file.\nAllowed label values for this entity are `L` and `R`, for the left and right\nhemibrains, respectively.\n',
                  'type': 'string', 'format': 'label', 'enum': ['L', 'R']}
    space = {'name': 'space', 'display_name': 'Space',
             'description': 'The `space-<label>` entity can be used to indicate the way in which electrode positions are interpreted\n(for EEG/MEG/iEEG data)\nor the spatial reference to which a file has been aligned (for MRI data).\nThe `<label>` MUST be taken from one of the modality specific lists in the\n[Coordinate Systems Appendix](SPEC_ROOT/appendices/coordinate-systems.md).\nFor example, for iEEG data, the restricted keywords listed under\n[iEEG Specific Coordinate Systems](SPEC_ROOT/appendices/coordinate-systems.md#ieeg-specific-coordinate-systems)\nare acceptable for `<label>`.\n\nFor EEG/MEG/iEEG data, this entity can be applied to raw data,\nbut for other data types, it is restricted to derivative data.\n',
             'type': 'string', 'format': 'label'}
    split = {'name': 'split', 'display_name': 'Split',
             'description': 'In the case of long data recordings that exceed a file size of 2Gb,\n`.fif` files are conventionally split into multiple parts.\nEach of these files has an internal pointer to the next file.\nThis is important when renaming these split recordings to the BIDS convention.\n\nInstead of a simple renaming, files should be read in and saved under their\nnew names with dedicated tools like [MNE-Python](https://mne.tools/),\nwhich will ensure that not only the filenames, but also the internal file pointers, will be updated.\n\nIt is RECOMMENDED that `.fif` files with multiple parts use the `split-<index>` entity to indicate each part.\nIf there are multiple parts of a recording and the optional `scans.tsv` is provided,\nall files MUST be listed separately in `scans.tsv` and\nthe entries for the `acq_time` column in `scans.tsv` MUST all be identical,\nas described in [Scans file](SPEC_ROOT/modality-agnostic-files.md#scans-file).\n',
             'type': 'string', 'format': 'index'}
    recording = {'name': 'recording', 'display_name': 'Recording',
                 'description': 'The `recording-<label>` entity can be used to distinguish continuous recording files.\n\nThis entity is commonly applied when continuous recordings have different sampling frequencies or start times.\nFor example, physiological recordings with different sampling frequencies may be distinguished using\nlabels like `recording-100Hz` and `recording-500Hz`.\n',
                 'type': 'string', 'format': 'label'}
    chunk = {'name': 'chunk', 'display_name': 'Chunk',
             'description': 'The `chunk-<index>` key/value pair is used to distinguish between images of\nthe same physical sample with different fields of view acquired in the same\nimaging experiment.\nThis entity applies to collections of 2D images, 3D volumes or 4D volume series\n(for example, diffusion weighted images), and may be used to indicate different\nanatomical structures or regions of the same structure.\n',
             'type': 'string', 'format': 'index'}
    segmentation = {'name': 'seg', 'display_name': 'Segmentation',
                    'description': 'The `seg-<label>` key/value pair corresponds to a custom label the user\nMAY use to distinguish different segmentations.\n\nThis entity is only applicable to derivative data.\n',
                    'type': 'string', 'format': 'label'}
    resolution = {'name': 'res', 'display_name': 'Resolution',
                  'description': 'Resolution of regularly sampled N-dimensional data.\n\nThis entity represents the `"Resolution"` metadata field.\nTherefore, if the `res-<label>` entity is present in a filename,\n`"Resolution"` MUST also be added in the JSON file, to provide interpretation.\n\nThis entity is only applicable to derivative data.\n',
                  'type': 'string', 'format': 'label'}
    density = {'name': 'den', 'display_name': 'Density',
               'description': 'Density of non-parametric surfaces.\n\nThis entity represents the `"Density"` metadata field.\nTherefore, if the `den-<label>` entity is present in a filename,\n`"Density"` MUST also be added in the JSON file, to provide interpretation.\n\nThis entity is only applicable to derivative data.\n',
               'type': 'string', 'format': 'label'}
    label = {'name': 'label', 'display_name': 'Label',
             'description': 'Tissue-type label, following a prescribed vocabulary.\nApplies to binary masks and probabilistic/partial volume segmentations\nthat describe a single tissue type.\n\nThis entity is only applicable to derivative data.\n',
             'type': 'string', 'format': 'label'}
    description = {'name': 'desc', 'display_name': 'Description',
                   'description': 'When necessary to distinguish two files that do not otherwise have a\ndistinguishing entity, the `desc-<label>` entity SHOULD be used.\n\nThis entity is only applicable to derivative data.\n',
                   'type': 'string', 'format': 'label'}



from ancpbids import load_dataset
from ..base_test_case import *


class DerivativesTestCase(BaseTestCase):
    def test_derivative_generated_by(self):
        test_ds = load_dataset(SYNTHETIC_DIR)
        schema = test_ds.get_schema()
        fmriprep_folder = test_ds.derivatives.get_folder('fmriprep')
        self.assertTrue(isinstance(fmriprep_folder, schema.DerivativeFolder))
        self.assertTrue(isinstance(fmriprep_folder.dataset_description, schema.DerivativeDatasetDescriptionFile))
        dddf = fmriprep_folder.dataset_description
        self.assertEqual(1, len(dddf.GeneratedBy))
        generated_by = dddf.GeneratedBy[0]
        self.assertEqual("fmriprep", generated_by.Name)
        self.assertEqual("1.1.0", generated_by.Version)
        self.assertTrue(generated_by.Container)
        container = generated_by.Container
        self.assertEqual("abc", container.Type)
        self.assertEqual("xyz", container.Tag)
        self.assertEqual("test:abc/xyz", container.URI)

    def test_derivative_dataset_description(self):
        test_ds = load_dataset(DS005_SMALL2_DIR)
        schema = test_ds.get_schema()
        dd_files = test_ds.select(schema.DatasetDescriptionFile).objects(as_list=True)
        self.assertEqual(2, len(dd_files))
        names = {'Mixed-gambles task', 'Mixed-gambles task -- dummy derivative'}
        dd_names = [d['Name'] for d in dd_files]
        self.assertTrue(set(dd_names) == names)

        dd = dd_files[1]
        # PipelineDescription is not part of BIDS spec but available in the generic contents object
        self.assertEqual('events', dd.contents['PipelineDescription']['Name'])

    def test_create_artifact_with_raw(self):
        test_ds = load_dataset(DS005_SMALL2_DIR)
        sub01_json = test_ds.query(sub='01', suffix='bold', extension='.json')[0]
        derivative_folder = test_ds.create_derivative(name="unit-test")
        deriv_artifact = derivative_folder.create_artifact(raw=sub01_json)
        deriv_artifact.add_entities(desc='unittest')
        self.assertEqual("01", deriv_artifact.get_entity("sub"))
        self.assertEqual("mixedgamblestask", deriv_artifact.get_entity("task"))
        self.assertEqual(1, deriv_artifact.get_entity("run"))
        self.assertEqual("unittest", deriv_artifact.get_entity("desc"))

if __name__ == '__main__':
    unittest.main()



import unittest

from ancpbids import load_dataset, DatasetOptions
from tests.base_test_case import DS005_DIR_IGNORED_RESOURCES


class IgnoredResourcesTestCase(unittest.TestCase):
    def test_ignore_file(self):
        ds = load_dataset(DS005_DIR_IGNORED_RESOURCES, DatasetOptions(ignore=True))
        assert ds.get_folder("models") is None
        assert ds.get_folder("stimuli") is None
        assert ds.get_file(".dummy") is None
        assert ds.query(sub="01", suffix="bold") is not None
        assert ds.dataset_description is not None

        assert ds.get_file(".bidsignore") is None

    def test_ignore_file_dynamic(self):
        ds = load_dataset(DS005_DIR_IGNORED_RESOURCES, DatasetOptions(ignore=["models", "stimuli", ".*"]))
        assert ds.get_folder("models") is None
        assert ds.get_folder("stimuli") is None
        assert ds.get_file(".dummy") is None
        assert ds.query(sub="01", suffix="bold") is not None
        assert ds.dataset_description is not None

        assert ds.get_file(".bidsignore") is None


if __name__ == '__main__':
    unittest.main()



from typing import List, Dict

import torch

from ancpbids.torch import TorchDataset
from ..base_test_case import *


class BasicTestCase(BaseTestCase):
    def test_torch_dataset(self):
        # load the dataset given the directory path, limiting to BOLD files (suffix="bold", extension='.nii.gz')
        torch_ds = TorchDataset(DS005_DIR, suffix="bold")
        train_set, test_set, val_set = torch_ds.split(0.8, 0.1, 0.1)

        assert len(train_set) == 40
        assert len(test_set) == 5
        assert len(val_set) == 4

        train_dl: torch.utils.data.DataLoader = torch.utils.data.DataLoader(train_set, shuffle=True)
        file_path, onset_events = next(iter(train_dl))
        assert os.path.exists(file_path[0])
        assert isinstance(onset_events, list)


if __name__ == '__main__':
    unittest.main()



from unittest import skip

import sys

from ..base_test_case import *

OPENNEURO_DS001734 = os.path.join(os.environ.get('TEST_DATASETS', os.path.expanduser('~/datasets')), 'ds001734')

if not os.path.isdir(OPENNEURO_DS001734):
    print('test dataset not found: ' + OPENNEURO_DS001734)
    sys.exit(1)


class BenchmarkTestCase(BaseTestCase):
    def _assert_on(self, layout_type):
        layout = layout_type(OPENNEURO_DS001734, derivatives=True)
        subjects = layout.get_subjects(scope="raw")
        self.assertEqual(108, len(subjects))
        self.assertEqual(0, len(layout.get_sessions()))

        bold_run1 = layout.get(scope="raw", suffix='bold', run='01', extension='.nii.gz', return_type='filename')
        self.assertEqual(108, len(bold_run1))


    def test_ancpbids_openneuro_ds001734(self):
        import ancpbids
        self._assert_on(ancpbids.BIDSLayout)

    def test_pybids_measure_scan_ds001734(self):
        import bids
        self._assert_on(bids.BIDSLayout)



# Simplified test case
import unittest
import time
import sys
from unittest.mock import patch, MagicMock
import os
import time
import unittest

import numpy as np
import pandas as pd
import shutil
import tempfile

from numpy.testing import tempdir

import ancpbids
from ancpbids import model_latest, re
from ..base_test_case import BaseTestCase, DS005_DIR
from ancpbids import DatasetOptions
from ancpbids import load_dataset, DatasetOptions



class LazyLoadingTestCase(BaseTestCase):
    """Test cases for lazy loading functionality."""

    def test_lazy_loading_prevents_immediate_loading(self):
        """Test that lazy loading prevents immediate content loading using mocking."""
        with patch('ancpbids.utils.load_contents') as mock_load:
            mock_load.return_value = {'test': 'data'}

            # Load with lazy loading enabled
            options = DatasetOptions(lazy_loading=True)
            ds005 = load_dataset(DS005_DIR, options)

            # Mock should not have been called during initialization
            # (or called much less frequently than with eager loading)
            lazy_call_count = mock_load.call_count

        with patch('ancpbids.utils.load_contents') as mock_load:
            mock_load.return_value = {'test': 'data'}

            # Load with lazy loading disabled
            options = DatasetOptions(lazy_loading=False)
            ds005_eager = load_dataset(DS005_DIR, options)

            eager_call_count = mock_load.call_count

        # Lazy loading should result in fewer immediate load calls
        # dataset_description.json are configured to not use lazy-loading, for example
        self.assertLess(lazy_call_count, eager_call_count)

    def test_lazy_loading_on_demand(self):
        """Test that lazy-loaded content can be accessed when needed."""
        options = DatasetOptions(lazy_loading=True)
        ds005 = load_dataset(DS005_DIR, options)

        # Get a lazy-loaded TSV file
        participants_file = ds005.get_file("participants.tsv")

        # Access contents - this should trigger loading
        contents = participants_file.contents
        self.assertIsNotNone(contents)
        self.assertEqual(16, len(contents))
        self.assertEqual(['participant_id', 'sex', 'age'], list(contents[0].keys()))

    def test_lazy_vs_eager_content_equivalence(self):
        """Test that lazy and eager loading produce equivalent results."""
        # Load with lazy loading
        options_lazy = DatasetOptions(lazy_loading=True)
        ds005_lazy = load_dataset(DS005_DIR, options_lazy)

        # Load with eager loading
        options_eager = DatasetOptions(lazy_loading=False)
        ds005_eager = load_dataset(DS005_DIR, options_eager)

        # Compare participants file contents
        lazy_participants = ds005_lazy.get_file("participants.tsv").contents
        eager_participants = ds005_eager.get_file("participants.tsv").contents

        self.assertEqual(len(lazy_participants), len(eager_participants))
        self.assertEqual(lazy_participants[0], eager_participants[0])

    def test_lazy_vs_eager_performance(self):
        """Test performance difference between lazy and eager loading."""
        measurements_lazy = []
        measurements_eager = []

        # Perform multiple measurements for stability
        for _ in range(3):
            # Time lazy loading
            start_time = time.time()
            options_lazy = DatasetOptions(lazy_loading=True)
            ds005_lazy = load_dataset(DS005_DIR, options_lazy)
            lazy_time = time.time() - start_time
            measurements_lazy.append(lazy_time)

            # Time eager loading
            start_time = time.time()
            options_eager = DatasetOptions(lazy_loading=False)
            ds005_eager = load_dataset(DS005_DIR, options_eager)
            eager_time = time.time() - start_time
            measurements_eager.append(eager_time)

        # Calculate means
        mean_lazy = sum(measurements_lazy) / len(measurements_lazy)
        mean_eager = sum(measurements_eager) / len(measurements_eager)

        # Lazy loading should generally be faster for initial loading
        print(f"Mean lazy loading time: {mean_lazy:.4f}s")
        print(f"Mean eager loading time: {mean_eager:.4f}s")

        # For small datasets, the difference might be minimal, so we just verify
        # that both complete successfully rather than asserting performance
        self.assertGreaterEqual(mean_lazy, 0)
        self.assertGreaterEqual(mean_eager, 0)

    def test_memory_usage_difference(self):
        """Test memory usage difference between lazy and eager loading."""
        import sys

        # Measure memory usage of lazy loading
        options_lazy = DatasetOptions(lazy_loading=True)
        ds005_lazy = load_dataset(DS005_DIR, options_lazy)
        lazy_size = sys.getsizeof(ds005_lazy)

        # Measure memory usage of eager loading
        options_eager = DatasetOptions(lazy_loading=False)
        ds005_eager = load_dataset(DS005_DIR, options_eager)
        eager_size = sys.getsizeof(ds005_eager)

        print(f"Lazy dataset object size: {lazy_size} bytes")
        print(f"Eager dataset object size: {eager_size} bytes")

        # This is a basic check - in practice, you might want to use more
        # sophisticated memory profiling tools
        self.assertGreaterEqual(lazy_size, 0)
        self.assertGreaterEqual(eager_size, 0)

    def test_lazy_loading_caching(self):
        """Test that lazy-loaded content is cached after first access."""
        options = DatasetOptions(lazy_loading=True)
        ds005 = load_dataset(DS005_DIR, options)

        participants_file = ds005.get_file("participants.tsv")

        with patch('ancpbids.utils.load_contents') as mock_load:
            mock_load.return_value = [{'test': 'data'}]

            # First access should trigger loading
            contents1 = participants_file.contents
            first_call_count = mock_load.call_count

            # Second access should use cached content
            contents2 = participants_file.contents
            second_call_count = mock_load.call_count

            # Should be the same object and no additional calls
            self.assertIs(contents1, contents2)
            self.assertEqual(first_call_count, second_call_count)


if __name__ == '__main__':
    unittest.main()


import os
import time
import unittest

import numpy as np
import pandas as pd
import shutil
import tempfile

from numpy.testing import tempdir

import ancpbids
from ancpbids import model_latest, re
from ..base_test_case import BaseTestCase, DS005_DIR


class WritingTestCase(BaseTestCase):

    def write_test_derivative(self):
        dataset = ancpbids.load_dataset(DS005_DIR)
        pipeline_name = "mypipeline-%d" % time.time()
        derivative = dataset.create_derivative(name=pipeline_name)
        derivative.dataset_description.GeneratedBy.Name = "My Test Pipeline"
        ents = dataset.query_entities()
        task_label = list(ents['task'])[0]

        for sub_label in ents["subject"]:
            subject = derivative.create_folder(name='sub-' + sub_label)
            # create an additional folder level to increase complexity of generation
            session = subject.create_folder(name='ses-01')

            # do some complex task
            # ... doing complex task ...
            # ... done
            txt_artifact = session.create_artifact()
            txt_artifact.add_entities(desc="mypipeline", run=1, sub=sub_label)
            txt_artifact.suffix = 'textual'
            txt_artifact.extension = ".txt"
            txt_artifact.content = "Subject %s participated in task %s" % (sub_label, task_label)

            # create some random data
            df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list('ABCD'))
            ev_artifact = session.create_artifact()
            ev_artifact.add_entities(desc="mypipeline", run=1, sub=sub_label)
            ev_artifact.suffix = 'events'
            ev_artifact.extension = ".tsv"
            # at this point, the file path is not known and will be provided
            # to lambda when the derivative is written to disk
            ev_artifact.content = lambda file_path, df=df: df.to_csv(file_path, index=None)

        ancpbids.write_derivative(dataset, derivative)
        return DS005_DIR, pipeline_name

    def test_write_derivative(self):
        # create a temporary dataset with a test derivative and return its root path and the created derivative
        ds_path, pipeline_name = self.write_test_derivative()
        # pretend loading a new dataset
        dataset = ancpbids.load_dataset(ds_path)
        # get the underlying graph/dataset for further inspection
        derivative_folder = filter(lambda f: f.name == pipeline_name, dataset.derivatives.folders)
        self.assertIsNotNone(derivative_folder)
        derivative_folder = next(derivative_folder)
        schema = dataset.get_schema()
        subjects = derivative_folder.select(schema.Folder) \
            .where(re(schema.Folder.name, r"sub-[\d]+")) \
            .objects(True)
        self.assertEqual(16, len(subjects))

        for i, subject in enumerate(subjects):
            sub_label = "%02d" % (i + 1)
            expected_sub_name = "sub-%s" % sub_label
            self.assertEqual(expected_sub_name, subject.name)

            expected_ses_name = "ses-01"
            self.assertEqual(expected_ses_name, subject.folders[0].name)

            ses_folder = subject.folders[0]
            for artifact in ses_folder.files:
                ents = artifact.get_entities()
                self.assertEqual(
                    "{'sub': '%s', 'ses': '01', 'run': 1, 'desc': 'mypipeline'}" % (sub_label),
                    str(ents))
                self.assertTrue(artifact.name.startswith('sub-%s_ses-01_run-1_desc-mypipeline_' % sub_label))

        self.assertTrue(isinstance(derivative_folder.dataset_description, schema.DerivativeDatasetDescriptionFile))
        self.assertEqual(derivative_folder.dataset_description.GeneratedBy.Name, "My Test Pipeline")
        shutil.rmtree(derivative_folder.get_absolute_path())

    def test_create_new_dataset(self):
        from ancpbids import model_latest as schema
        dataset = schema.create_dataset(name='my-test-ds')
        dataset.dataset_description.Name = 'a programmatically created dataset'
        dataset.dataset_description.BIDSVersion = schema.VERSION

        for i in range(1, 10):
            subject = dataset.create_folder(name=f"sub-{i}", type_=schema.Subject)
            func_folder = subject.create_folder(name='func', type_=schema.DatatypeFolder)
            img_file = func_folder.create_artifact()
            img_file.suffix = 'bold'
            img_file.extension = '.nii.gz'
            img_file.add_entity('task', 'programming')
            # just create an empty file
            img_file.content = lambda file_name: open(file_name, 'w').close()

        output_dir = tempfile.mkdtemp()
        ancpbids.save_dataset(dataset, output_dir)

        # TODO reload dataset and add assertions

    def test_write_intermediate_results(self):
        from ancpbids import model_latest as schema
        output_dir = tempfile.mkdtemp()
        dataset = schema.create_dataset(output_dir, name='my-test-ds')
        dataset.dataset_description.Name = 'a programmatically created dataset'
        dataset.dataset_description.BIDSVersion = schema.VERSION

        for i in range(1, 10):
            subject = dataset.create_folder(name=f"sub-{i}", type_=schema.Subject)
            func_folder = subject.create_folder(name='func', type_=schema.DatatypeFolder)
            img_file = func_folder.create_artifact()
            img_file.suffix = 'bold'
            img_file.extension = '.nii.gz'
            img_file.add_entity('task', 'programming')
            # content callback just creates an empty file
            img_file.content = lambda file_name: open(file_name, 'w').close()
            # immediately create the file
            file_path: str = img_file.write()
            # we only care about the path beginning with the derivative folder
            expected_relative_file_path = os.path.normpath(
                f"my-test-ds/sub-{i}/func/sub-{i}_task-programming_bold.nii.gz")
            self.assertTrue(file_path.endswith(expected_relative_file_path))

            # we can also get its absolute path name
            file_path_abs = img_file.get_absolute_path()
            self.assertEqual(file_path_abs, file_path, "Expected touch to return the absolute path of the file")
            self.assertTrue(os.path.exists(file_path), f"Missing expected file {file_path}")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys, os
from setuptools import setup
import versioneer

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = ['setuptools >= 30.3.0']
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []
INSTALL_REQUIRES = []

if __name__ == '__main__':
    REQ_FILE_PATH = os.path.dirname(os.path.realpath(__file__)) + "/requirements.txt"
    with open(REQ_FILE_PATH) as f:
        INSTALL_REQUIRES = list(f.read().splitlines())
    setup(name="ancpbids",
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass(),
          setup_requires=SETUP_REQUIRES,
          install_requires=INSTALL_REQUIRES)
